import datetime as dt
from functools import lru_cache, partial
import json
import logging


import pandas as pd
from requests.exceptions import HTTPError


from solarforecastarbiter.datamodel import Observation, ProbabilisticForecast
from solarforecastarbiter.io.reference_observations.default_forecasts import (
    CURRENT_NWP_VARIABLES, is_in_nwp_domain)
from solarforecastarbiter.reference_forecasts.utils import (
    check_persistence_compatibility)


logger = logging.getLogger('reference_data')


def decode_extra_parameters(metadata):
    """Returns a dictionary parsed from the json string stored
    in extra_parameters

    Parameters
    ----------
    metadata
        A SolarForecastArbiter.datamodel class with an extra_parameters
        attribute

    Returns
    -------
    dict
        The extra parameters as a python dictionary

    Raises
    ------
    ValueError
        If parameters cannot be decoded or are None. Or if missing the
        required keys: network, network_api_id, network_api_abbreviation
        and observation_interval_length.
    """
    try:
        params = json.loads(metadata.extra_parameters)
    except (json.decoder.JSONDecodeError, TypeError):
        raise ValueError(f'Could not read extra parameters of {metadata.name}')
    required_keys = ['network', 'network_api_id', 'network_api_abbreviation',
                     'observation_interval_length']
    if not all([key in params for key in required_keys]):
        raise ValueError(f'{metadata.name} is missing required extra '
                         'parameters.')
    return params


def check_network(networks, metadata):
    """Decodes extra_parameters and checks if an object is in
    the network.

    Parameters
    ----------
    networks: list of str
        A list of networks to check against.
    metadata
        An instantiated dataclass from the datamodel.

    Returns
    -------
    bool
        True if the site belongs to the network, or one of the
        networks.
    """
    if type(networks) == str:
        networks = [networks]
    try:
        extra_params = decode_extra_parameters(metadata)
    except ValueError:
        return False
    try:
        in_network = extra_params['network'] in networks
    except KeyError:
        return False
    else:
        return in_network


@lru_cache(maxsize=4)
def existing_observations(api):
    return {obs.name: obs for obs in api.list_observations()}


@lru_cache(maxsize=4)
def existing_forecasts(api):
    out = {fx.name: fx for fx in api.list_forecasts()}
    out.update({fx.name: fx for fx in api.list_probabilistic_forecasts()})
    return out


@lru_cache(maxsize=4)
def existing_sites(api):
    return {site.name: site for site in api.list_sites()}


def filter_by_networks(object_list, networks):
    """Returns a copy of object_list with all objects that are not in the
    network removed.

    Parameters
    ----------
    object_list: list
        List of datamodel objects.
    networks: string or list
        Network or list of networks to check for.

    Returns
    -------
    filtered
        List of filtered datamodel objects.
    """
    filtered = [obj for obj in object_list if check_network(networks, obj)]
    return filtered


def create_observation(api, site, variable, extra_params=None, **kwargs):
    """ Creates a new Observation for the variable and site. Kwargs can be
    provided to overwrite the default arguments to the Observation constructor.
    Kwarg options are documented in 'Other Parameters' below but users should
    reference the SolarForecastArbiter API for valid Observation field values.

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An APISession with a valid JWT for accessing the Reference Data user.
    site : solarforecastarbiter.datamodel.site
        A site object.
    variable : string
        Variable measured in the observation.
    extra_params : dict, optional
        If provided, this dict will be serialized as the 'extra_parameters'
        field of the observation, otherwise the site's field is copied over.
        Must contain the keys 'network_api_length', 'network_api_id', and
        'observation_interval_length'.

    Other Parameters
    ----------------
    name: string
        Defaults to `<site.name> <variable>`
    interval_label: string
        Defaults to 'ending'
    interval_value_type: string
        Defaults to 'interval_mean'
    uncertainty: float
        Defaults to 0.

    Returns
    -------
    created
        The datamodel object of the newly created observation.

    Raises
    ------
    KeyError
        When the extra_parameters, either loaded from the site or provided
        by the user is missing 'network_api_abbreviation'
        or 'observation_interval_length'

    """
    # Copy network api data from the site, and get the observation's
    # interval length
    if extra_params:
        extra_parameters = extra_params
    else:
        try:
            extra_parameters = decode_extra_parameters(site)
        except ValueError:
            logger.warning(f'Cannot create observations for site {site.name}'
                           'missing required extra parameters')
            return
    site_name = site_name_no_network(site)
    observation_name = f'{site_name} {variable}'
    # Some site names are too long and exceed the API's limits,
    # in those cases. Use the abbreviated version.
    if len(observation_name) > 64:
        site_abbreviation = extra_parameters["network_api_abbreviation"]
        observation_name = f'{site_abbreviation} {variable}'

    observation = Observation.from_dict({
        'name': kwargs.get('name', observation_name),
        'interval_label': kwargs.get('interval_label', 'ending'),
        'interval_length': extra_parameters['observation_interval_length'],
        'interval_value_type': kwargs.get('interval_value_type',
                                          'interval_mean'),
        'site': site,
        'uncertainty': kwargs.get('uncertainty', 0),
        'variable': variable,
        'extra_parameters': json.dumps(extra_parameters)
    })

    return check_and_post_observation(api, observation)


def check_and_post_observation(api, observation):
    existing = existing_observations(api)
    if observation.name in existing:
        logger.info('Observation, %s, already exists', observation.name)
        return existing[observation.name]

    try:
        created = api.create_observation(observation)
    except HTTPError as e:
        logger.error(f'Failed to create {observation.variable} observation '
                     f'at Site {observation.site.name}.')
        logger.debug(f'HTTP Error: {e.response.text}')
    else:
        logger.info(f"Observation {created.name} created successfully.")
        return created


def _utcnow():
    return pd.Timestamp.now(tz='UTC')


def get_last_site_timestamp(api, observations, end):
    """Get the last value timestamp from the API as the minimum of the
    last timestamp for each observation at that site. The result of
    this function is often used to make new data queries, so a limit
    of end - 7 days is set to avoid excessive queries to external
    sources.

    Parameters
    ---------
    api : solarforecastarbiter.io.api.APISession
        An active Reference user session.
    site_observations : list of solarforecastarbiter.datamodel.Observation
        A list of reference Observations for a site to search.
    end : pd.Timestamp
        Typically, set to now.

    Returns
    -------
    pandas.Timestamp

    """
    # update cli.py as appropriate if behaviour is changed
    out = end
    updated = False
    for obs in observations:
        maxt = api.get_observation_time_range(obs.observation_id)[1]
        # <= so that even if maxt == end updated -> true
        # effectively ignores all NaT values unless all observations
        # for the site are NaT, then use weekago
        if pd.notna(maxt) and maxt <= out:
            out = maxt
            updated = True

    weekago = end - pd.Timedelta('7d')
    if not updated or out < weekago:
        out = weekago
    return out


def update_site_observations(api, fetch_func, site, observations,
                             start, end):
    """Updates data for all reference observations at a given site
    for the period between start and end.

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An active Reference user session.
    fetch_func : function
        A function that requests data and returns a DataFrame for a given site.
        The function should accept the parameters (api, site, start, end) as
        they appear in this function.
    site : solarforecastarbiter.datamodel.Site
        The Site with observations to update.
    observations : list of solarforecastarbiter.datamodel.Observation
        A full list of reference Observations to search.
    start : pandas.Timestamp or None
        Start time to get data for. If None, try finding the last
        value in the API and use that time (with a limit of 7 days from start
        to end). If None and no values in the API, use end - 7 day.
    end : pandas.Timestamp or None
        End time to get data for. If None, use now.
    """
    site_observations = [obs for obs in observations if obs.site == site]
    if end is None:
        end = _utcnow()
    if start is None:
        start = get_last_site_timestamp(api, site_observations, end)
    logger.debug('Fetching data for %s from %s to %s', site.name, start, end)
    obs_df = fetch_func(api, site, start, end)
    # must be sorted for proper inexact start:end slicing
    data_in_range = obs_df.sort_index()[start:end]
    if data_in_range.empty:
        return
    for obs in site_observations:
        post_observation_data(api, obs, data_in_range, start, end)


def _prepare_data_to_post(data, variable, observation, start, end,
                          resample_how=None):
    """Manipulate the data including reindexing to observation.interval_label
    to prepare for posting"""
    data = data[[variable]]
    data = data.rename(columns={variable: 'value'})
    # ensure data is sorted before slicing and for optimal order in the
    # database
    data = data.sort_index()

    if resample_how:
        resampler = data.resample(observation.interval_length)
        data = getattr(resampler, resample_how)()

    # remove all future values, some files have forward filled nightly data
    data = data[start:min(end, _utcnow())]

    if data.empty:
        return data
    # reindex the data to put nans where required
    # we don't extend the new index to start, end, since reference
    # data has some lag time from the end it was requested from
    # and it isn't necessary to keep the nans between uploads in db
    new_index = pd.date_range(start=data.index[0], end=data.index[-1],
                              freq=observation.interval_length)
    data = data.reindex(new_index)
    # set quality flags
    data['quality_flag'] = data['value'].isna().astype(int)
    return data


def post_observation_data(api, observation, data, start, end):
    """Posts data to an observation between start and end.

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An active Reference user session.
    observation : solarforecastarbiter.datamodel.Observation
        Data model object corresponding to the Observation to update.
    data : pandas.DataFrame
        Dataframe of values to post containing a column labeled with
        the Observation's variable.
    start : datetime-like
        The beginning of the period to update.
    end : datetime-like
        The end of the period to update.

    Raises
    ------
    IndexError
        If the data provided has an empty index.
    """
    logger.info(
        f'Updating {observation.name} from '
        f'{start} to {end}.')
    try:
        extra_parameters = decode_extra_parameters(observation)
    except ValueError:
        return
    # check for a non-standard variable label in extra_parameters
    variable = extra_parameters.get('network_data_label',
                                    observation.variable)
    # check if the raw observation needs to be resampled before posting
    resample_how = extra_parameters.get('resample_how', None)
    try:
        var_df = _prepare_data_to_post(data, variable, observation,
                                       start, end, resample_how=resample_how)
    except KeyError:
        logger.error(f'{variable} could not be found in the data file '
                     f'from {data.index[0]} to {data.index[-1]}'
                     f'for Observation {observation.name}')
        return
    except AttributeError:
        logger.error(f'{variable} could not be resampled using method '
                     f'{resample_how} for Observation {observation.name}')
        return

    # skip post id data is empty, if there are nans, should still post
    if var_df.empty:
        logger.warning(
            f'{observation.name} data empty from '
            f'{data.index[0]} to {data.index[-1]}.')
        return
    try:
        logger.debug(f'Posting data to {observation.name} between '
                     f'{var_df.index[0]} and {var_df.index[-1]}.')
        api.post_observation_values(observation.observation_id, var_df)
    except HTTPError as e:
        logger.error(f'Posting data to {observation.name} failed.')
        logger.debug(f'HTTP Error: {e.response.text}.')


def clean_name(string):
    """Removes all disallowed characters from a string and converts
    underscores to spaces.
    """
    return string.translate(string.maketrans('_', ' ', ':(){}/\\[]@-.'))


def site_name_no_network(site):
    """Removes the prefixed network from a site name for prepending
    to an observation.
    """
    extra_params = decode_extra_parameters(site)
    network = extra_params['network']
    # only select the site name after the network name and a space.
    if site.name.startswith(network):
        return site.name[len(network) + 1:]
    else:
        return site.name


def _make_fx_name(site_name, template_name, variable):
    fx_name = f'{site_name} {template_name} {variable}'
    # Some site names are too long and exceed the API's limits,
    # in those cases. Use the abbreviated version.
    if len(fx_name) > 63:
        for old, new in (
            ('Persistence', 'Pers'),
            ('persistence', 'pers'),
            ('Fifteen-minute', '15 min'),
            ('Five-minute', '5 min'),
        ):
            template_name = template_name.replace(old, new)
        suffix = f'{template_name} {variable}'
        # ensure name can containe at least first word of site
        # name and the suffix
        if (len(site_name.split(' ')[0]) + len(suffix)) > 62:
            raise ValueError('Template/site name too long together')
        while len(fx_name) > 63:
            # drop the last word
            site_name = ' '.join(site_name.split(' ')[:-1])
            fx_name = f"{site_name} {suffix}"
        logger.warning("Forecast name truncated to %s", fx_name)
    return fx_name


def create_one_forecast(api, site, template_forecast, variable,
                        creation_validation=lambda x: True, **extra_params):
    """Creates a new Forecast or ProbabilisticForecast for the variable
    and site based on the template forecast.

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An APISession with a valid JWT for accessing the Reference Data user.
    site : solarforecastarbiter.datamodel.site
        A site object.
    template_forecast : solarforecastarbiter.datamodel.Forecast
        A Forecast or ProbabilisticForecast object that will only have name,
        site, variable, and issue_time_of_day replaced. New keys may be added
        to extra parameters.
    variable : string
        Variable measured in the forecast.
    creation_validation : function
        Function that expects a Forecast or ProbabilisticForecast object
        and raises a ValueError if the forecast is invalid just before it
        is created.
    **extra_params : dict
        Other key, value pairs to add to the extra_parameters of the Forecast
        object.

    Returns
    -------
    created
        The datamodel object of the newly created forecast.
    """
    extra_parameters = json.loads(template_forecast.extra_parameters)
    extra_parameters.update(extra_params)
    site_name = site_name_no_network(site)
    fx_name = _make_fx_name(site_name, template_forecast.name, variable)
    # adjust issue_time_of_day to localtime for standard time, not DST
    issue_datetime = pd.Timestamp.combine(
        dt.date(2019, 2, 1), template_forecast.issue_time_of_day,
        ).tz_localize(site.timezone).tz_convert('UTC')
    # make sure this is the first possible issue for the UTC day
    orig_date = issue_datetime.floor('1d')
    while issue_datetime - template_forecast.run_length >= orig_date:
        issue_datetime -= template_forecast.run_length
    issue_time_of_day = issue_datetime.time()

    forecast = template_forecast.replace(
        name=fx_name, extra_parameters=json.dumps(extra_parameters),
        site=site, variable=variable, issue_time_of_day=issue_time_of_day,
    )
    existing = existing_forecasts(api)
    if (
            forecast.name in existing and
            existing[forecast.name].site == forecast.site
    ):
        logger.info('Forecast, %s, already exists', forecast.name)
        return existing[forecast.name]

    if isinstance(forecast, ProbabilisticForecast):
        create_func = api.create_probabilistic_forecast
    else:
        create_func = api.create_forecast

    try:
        creation_validation(forecast)
    except ValueError as exc:
        logger.error('Validation failed on creation of %s forecast '
                     'at Site %s with message %s', variable, site.name, exc)
        return
    try:
        created = create_func(forecast)
    except HTTPError as e:
        logger.error(f'Failed to create {variable} forecast at Site '
                     f'{site.name}.')
        logger.debug(f'HTTP Error: {e.response.text}')
    else:
        logger.info(f"Forecast {created.name} created successfully.")
        return created


def create_nwp_forecasts(api, site, variables, templates):
    """Create Forecast objects for each of variables, if NWP forecasts
    can be made for that variable. Each Forecast in templates will be
    updated with the appropriate parameters for each variable. Forecasts
    will also be grouped together via 'piggyback_on'.

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An APISession with a valid JWT for accessing the Reference Data user.
    site : solarforecastarbiter.datamodel.Site
        A site object.
    variables : list-like
        List of variables to make a new forecast for each of the template
        forecasts
    templates : list of datamodel.Forecasts or datamodel.ProbabilisticForecast
        Forecasts that will be used as templates for many fields. See
        :py:func:`solarforecastarbiter.io.reference_data.common.create_one_forecast`
        for the fields that are required vs overwritten.

    Raises
    ------
    ValueError
        If the site is outside the domain of the current NWP forecasts.
    """  # NOQA
    if not is_in_nwp_domain(site):
        raise ValueError(
            f'Site {site.name} is outside the domain of the current NWP '
            'forecasts')
    vars_ = set(variables)
    diff = vars_ - CURRENT_NWP_VARIABLES
    if diff:
        logger.warning('NWP forecasts for %s cannot currently be made',
                       diff)
    vars_ = vars_ & CURRENT_NWP_VARIABLES
    if len(vars_) == 0:
        return []

    if 'ac_power' in vars_:
        primary = 'ac_power'
        vars_.remove('ac_power')
    elif 'ghi' in vars_:
        primary = 'ghi'
        vars_.remove('ghi')
    else:
        # pick random var
        primary = vars_.pop()

    created = []
    for template_fx in templates:
        logger.info('Creating forecasts based on %s at site %s',
                    template_fx.name, site.name)
        primary_fx = create_one_forecast(api, site, template_fx, primary)
        created.append(primary_fx)
        piggyback_on = primary_fx.forecast_id
        for var in vars_:
            created.append(
                create_one_forecast(api, site, template_fx, var,
                                    piggyback_on=piggyback_on))
    return created


def create_persistence_forecasts(api, site, variables, templates):
    """Create persistence Forecast objects for each Observation at the
    ``site`` with variable in ``variables``. Each Forecast in templates
     will be updated with the appropriate parameters for each variable.
    By default, *index* persistence forecasts are made for variables
    with valid index persistence functions namely (ghi, dni, dhi, ac_power).

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An APISession with a valid JWT for accessing the Reference Data user.
    site : solarforecastarbiter.datamodel.Site
        A site object with Observations whose data will be persisted.
    variables : list-like
        List of variables to make a new forecast for each of the template
        forecasts
    templates : list of datamodel.Forecasts or datamodel.ProbabilisticForecast
        Forecasts that will be used as templates for many fields. See
        :py:func:`solarforecastarbiter.io.reference_data.common.create_one_forecast`
        for the fields that are required vs overwritten.

    """  # NOQA
    created = []
    for obs in api.list_observations():
        if obs.site != site or obs.variable not in variables:
            continue
        for template_fx in templates:
            logger.info('Creating forecast based on %s and observation %s',
                        template_fx.name, obs.name)
            use_index = (
                template_fx.run_length < pd.Timedelta('1d') and
                obs.variable in ('ghi', 'dni', 'dhi', 'ac_power')
            )
            validation_func = partial(check_persistence_compatibility, obs,
                                      index=use_index)
            # net_load might go here, although other changes might be required
            fx_id = create_one_forecast(api, site, template_fx, obs.variable,
                                        creation_validation=validation_func,
                                        observation_id=obs.observation_id,
                                        index_persistence=use_index)
            created.append(fx_id)
    return created


def create_forecasts(api, site, variables, templates):
    """Create Forecast objects (NWP based and persistence) for each of
    variables. Each Forecast in templates will be updated with the
    appropriate parameters for each variable.

    Templates with the 'is_reference_persistence_forecast' key in
    'extra_parameters' are assumed to be persistence forecasts, and others
    are assumed to be NWP forecasts.

    Parameters
    ----------
    api : solarforecastarbiter.io.api.APISession
        An APISession with a valid JWT for accessing the Reference Data user.
    site : solarforecastarbiter.datamodel.Site
        A site object.
    variables : list-like
        List of variables to make a new forecast for each of the template
        forecasts
    templates : list of datamodel.Forecasts or datamodel.ProbabilisticForecast
        Forecasts that will be used as templates for many fields. See
        :py:func:`solarforecastarbiter.io.reference_data.common.create_one_forecast`
        for the fields that are required vs overwritten.

    Returns
    -------
    list
        A list of all Forecast/ProbabilisticForecast objects created

    See Also
    --------
    solarforecastarbiter.io.reference_data.common.create_nwp_forecasts
    solarforecastarbiter.io.reference_data.common.create_persistence_forecasts

    """  # NOQA: E501
    persistence_templates = []
    nwp_templates = []
    for template in templates:
        if 'is_reference_persistence_forecast' in template.extra_parameters:
            persistence_templates.append(template)
        else:
            nwp_templates.append(template)
    nwp_created = create_nwp_forecasts(api, site, variables, nwp_templates)
    persist_created = create_persistence_forecasts(
        api, site, variables, persistence_templates)
    return nwp_created + persist_created


def apply_json_site_parameters(json_sitefile, site):
    """Updates site metadata with modeling parameters found in a json file.

    Parameters
    ----------
    json_sitefile: str
        Absolute path of a json file with a 'sites' key containing a list of
        sites in the Solar Forecast Arbiter JSON format.
    site: dict

    Returns
    -------
    dict
        Copy of inputs plus a new key 'modeling_parameters'.
    """
    with open(json_sitefile) as fp:
        sites_metadata = json.load(fp)['sites']
    site_api_id = str(site['extra_parameters']['network_api_id'])
    for site_metadata in sites_metadata:
        site_extra_params = json.loads(site_metadata['extra_parameters'])
        if str(site_extra_params['network_api_id']) == site_api_id:
            site_out = site.copy()
            site_out['modeling_parameters'] = site_metadata[
                'modeling_parameters']
            site_out['extra_parameters'].update(site_extra_params)
            return site_out
    return site



import datetime as dt
from functools import partial
from contextlib import nullcontext as does_not_raise


import pandas as pd
from pandas.testing import assert_series_equal
import pytest
import numpy as np
from numpy.testing import assert_allclose
from scipy.stats import PearsonRConstantInputWarning


from solarforecastarbiter import datamodel
from solarforecastarbiter.metrics import deterministic


@pytest.fixture
def error_func_deadband():
    return partial(deterministic.error_deadband, deadband=0.05)


@pytest.mark.parametrize('deadband,expected', [
    (0., [1, 0, 0, 0, 0]),
    (0.1, [1, 0, 0, 0, 1]),
    (1., [1, 0, 1, 0, 1]),
])
def test_deadband_mask(deadband, expected):
    obs = np.array([0, 0, 1, 0, 1.])
    fx = np.array([0, 1, 0, 1.05, 0.95])
    expected = np.array(expected, dtype=bool)
    out = deterministic.deadband_mask(obs, fx, deadband)
    assert_allclose(out, expected)


def test_error():
    obs = np.array([2, 1, 0.])
    fx = np.array([1, 2, 0.])
    expected = np.array([-1, 1, 0.])
    out = deterministic.error(obs, fx)
    assert_allclose(out, expected)


@pytest.mark.parametrize('deadband,expected', [
    (0., [0, 1, -1, 1.05, -0.05]),
    (0.1, [0, 1, -1, 1.05, 0]),
    (1., [0, 1, 0, 1.05, 0]),
])
def test_error_deadband(deadband, expected):
    obs = np.array([0, 0, 1, 0, 1.])
    fx = np.array([0, 1, 0, 1.05, 0.95])
    expected = np.array(expected)
    out = deterministic.error_deadband(obs, fx, deadband)
    assert_allclose(out, expected)


@pytest.mark.parametrize("obs,fx,value", [
    (np.array([0, 1, 2]), np.array([0, 1, 2]), 0.0),
    (np.array([0, 1, 2]), np.array([0, 1, 1]), 1 / 3),
    (np.array([0, 1, 2]), np.array([0, 1, 3]), 1 / 3),
])
def test_mae(obs, fx, value):
    mae = deterministic.mean_absolute(obs, fx)
    assert mae == value


@pytest.mark.parametrize("obs,fx,value", [
    (np.array([0, 1, 2]), np.array([0, 1, 2]), 0.0),
    (np.array([0, 1, 2]), np.array([1, 0, 2]), 0.0),
    (np.array([0, 1, 2]), np.array([1, 2, 3]), 1.0),
    (np.array([0, 1, 2]), np.array([1, 3, 4]), (1 + 2 + 2) / 3),
    (np.array([5, 5, 5]), np.array([4, 4, 4]), -1.0),
    (np.array([5, 5, 5]), np.array([4, 3, 3]), -(1 + 2 + 2) / 3),
])
def test_mbe(obs, fx, value):
    mbe = deterministic.mean_bias(obs, fx)
    assert mbe == value


@pytest.mark.parametrize("obs,fx,value", [
    (np.array([0, 1]), np.array([0, 1]), 0.0),
    (np.array([0, 1]), np.array([1, 2]), 1.0),
    (np.array([1, 2]), np.array([0, 1]), 1.0),
])
def test_rmse(obs, fx, value):
    rmse = deterministic.root_mean_square(obs, fx)
    assert rmse == value


@pytest.mark.parametrize("obs,fx,value", [
    (np.array([1, 1]), np.array([2, 2]), 100.0),
    (np.array([2, 2]), np.array([3, 3]), 50.0),
    (np.array([1, 2]), np.array([1, 2]), 0.0),
])
def test_mape(obs, fx, value):
    mape = deterministic.mean_absolute_percentage(obs, fx)
    assert mape == value


@pytest.mark.parametrize("obs,fx,norm,value", [
    (np.array([0, 1, 2]), np.array([0, 1, 2]), 55, 0.0),
    (np.array([0, 1, 2]), np.array([0, 1, 1]), 20, 1 / 3 / 20 * 100),
])
def test_nmae(obs, fx, norm, value):
    nmae = deterministic.normalized_mean_absolute(obs, fx, norm)
    assert nmae == value


@pytest.mark.parametrize("obs,fx,norm,value", [
    (np.array([0, 1, 2]), np.array([0, 1, 2]), 55, 0.0),
    (np.array([0, 1, 2]), np.array([1, 0, 2]), 20, 0.0),
    (np.array([0, 1, 2]), np.array([1, 3, 4]), 7, (1 + 2 + 2) / 3 / 7 * 100),
    (np.array([5, 5, 5]), np.array([4, 4, 4]), 2, -1.0 / 2 * 100),
    (np.array([5, 5, 5]), np.array([4, 3, 3]), 2, -(1 + 2 + 2) / 3 / 2 * 100),
])
def test_nmbe(obs, fx, norm, value):
    nmbe = deterministic.normalized_mean_bias(obs, fx, norm)
    assert nmbe == value


@pytest.mark.parametrize("obs,fx,norm,value", [
    (np.array([0, 1, 2]), np.array([0, 1, 2]), 1.0, 0.0),
    (np.array([0, 1, 2]), np.array([0, 1, 2]), 55.0, 0.0),
    (np.array([0, 1]), np.array([1, 2]), 1.0, 100.0),
    (np.array([0, 1]), np.array([1, 2]), 100.0, 1.0),
])
def test_nrmse(obs, fx, norm, value):
    nrmse = deterministic.normalized_root_mean_square(obs, fx, norm)
    assert nrmse == value


@pytest.mark.parametrize("obs,fx,ref,value", [
    (np.array([0, 1]), np.array([0, 2]), np.array([0, 1]), np.NINF),
    (np.array([0, 1]), np.array([0, 1]), np.array([0, 1]), 0.0),
    (np.array([0, 1]), np.array([0, 2]), np.array([0, 2]), 0.0),
    (np.array([0, 1]), np.array([0, 2]), np.array([0, 3]), 0.5),
])
def test_skill(obs, fx, ref, value):
    s = deterministic.forecast_skill(obs, fx, ref)
    assert s == value


@pytest.mark.parametrize("obs,fx,value", [
    (np.array([0, 1]), np.array([0, 1]), 1.0),
    (np.array([1, 2]), np.array([-1, -2]), -1.0),
])
def test_r(obs, fx, value):
    r = deterministic.pearson_correlation_coeff(obs, fx)
    assert r == value


@pytest.mark.parametrize("obs,fx,context", [
    # len(obs) < 2 or len(fx) < 2
    (np.array([0]), np.array([1]), does_not_raise()),

    # len(obs) != len(fx)
    (np.array([0, 1, 2]), np.array([0, 1, 2, 3]), does_not_raise()),
    (np.array([2, 3, 4]), np.array([2, 3, 5, 6]), does_not_raise()),

    # obs or fx have the same values
    (np.array([1, 1, 1]), np.array([1, 2, 3]),
     pytest.warns(PearsonRConstantInputWarning)),
    (np.array([1, 2, 3]), np.array([1, 1, 1]),
     pytest.warns(PearsonRConstantInputWarning)),
])
def test_r_nan(obs, fx, context):
    with context:
        r = deterministic.pearson_correlation_coeff(obs, fx)
    assert np.isnan(r)


@pytest.mark.parametrize("obs,fx,value", [
    (np.array([0, 1]), np.array([0, 1]), 1.0),
    (np.array([1, 2, 3]), np.array([2, 2, 2]), 0.0),
])
def test_r2(obs, fx, value):
    r2 = deterministic.coeff_determination(obs, fx)
    assert pytest.approx(r2) == value


@pytest.mark.parametrize("obs,fx,value", [
    (np.array([0, 1]), np.array([0, 1]), 0.0),
    (np.array([0, 2]), np.array([0, 4]), 1.0),
    (np.array([0, 2]), np.array([0, 6]), 2.0),
])
def test_crmse(obs, fx, value):
    crmse = deterministic.centered_root_mean_square(obs, fx)
    assert crmse == value


@pytest.mark.parametrize("obs,fx,value,context", [
    (np.array([0, 1]), np.array([0, 1]), 0.0, does_not_raise()),
    (np.array([0, 1]), np.array([1, 2]), 2.0, does_not_raise()),
    (np.array([1, 2]), np.array([0, 1]), 0.6666666666666667, does_not_raise()),
    (np.array([-1, 1]), np.array([2, 3]), np.Inf, does_not_raise()),
    (np.array([-2, 2]), np.array([3, -3]), 2.0615528128088303,
     does_not_raise()),
    (np.array([1, 1]), np.array([2, 3]), np.nan,
     pytest.warns(PearsonRConstantInputWarning)),  # corr = nan
    (np.array([2, 3]), np.array([1, 1]), np.nan,
     pytest.warns(PearsonRConstantInputWarning)),  # corr = nan
    (np.array([0, 0]), np.array([1, 2]), np.nan,
     pytest.warns(PearsonRConstantInputWarning)),  # corr = nan
])
def test_relative_euclidean_distance(obs, fx, value, context):
    with context:
        out = deterministic.relative_euclidean_distance(obs, fx)
    if np.isnan(value):
        assert np.isnan(out)
    else:
        assert out == value


@pytest.mark.parametrize("obs,fx,value", [
    ([0, 1], [0, 1], 0.0),
    ([1, 2], [1, 2], 0.0),
    ([0, 1], [0, 2], 0.5),
    ([0, 1, 2], [0, 0, 2], 1.0 / 3.0),
])
def test_ksi(obs, fx, value):
    ksi = deterministic.kolmogorov_smirnov_integral(obs, fx)
    assert pytest.approx(ksi) == value


@pytest.mark.parametrize("obs,fx,value", [
    ([0, 1], [0, 1], 0.0),
    ([1, 2], [1, 2], 0.0),
    ([0, 1, 2], [0, 0, 2], 1 / 3 / (1.63 / np.sqrt(3) * 2) * 100),
])
def test_ksi_norm(obs, fx, value):
    ksi = deterministic.kolmogorov_smirnov_integral(
        obs, fx, normed=True
    )
    assert pytest.approx(ksi) == value


@pytest.mark.parametrize("obs,fx,value", [
    ([0, 1], [0, 1], 0.0),
    ([1, 2], [1, 2], 0.0),
    ([0, 1, 2, 3, 4], [0, 0, 0, 0, 0], 0.8 - 1.63 / np.sqrt(5)),
])
def test_over(obs, fx, value):
    ov = deterministic.over(obs, fx)
    assert ov == value


@pytest.mark.parametrize("obs,fx,value", [
    (np.array([0, 1]), np.array([0, 1]), 0.0),
    (np.array([1, 2]), np.array([1, 2]), 0.0),
    (
        np.array([0, 1, 2]),
        np.array([0, 0, 2]),
        1/4 * (1/3 + 0 + 2 * np.sqrt(1/3))
    ),
])
def test_cpi(obs, fx, value):
    cpi = deterministic.combined_performance_index(obs, fx)
    assert pytest.approx(cpi) == value


@pytest.fixture
def deadband_obs_fx():
    obs = np.array([1, 2, 3, 4])
    # 2.1 and 3.8 are outside the 5% deadband on some platforms due to
    # floating point arithmetic errors
    fx = np.array([2, 2.09, 2, 3.81])
    return obs, fx


@pytest.mark.parametrize('func,expect,expect_deadband,args', [
    (deterministic.mean_absolute, 0.57, 0.5, []),
    (deterministic.mean_bias, -0.025, 0., []),
    (deterministic.root_mean_square,
     0.7148776119029046, 0.7071067811865476, []),
    (deterministic.mean_absolute_percentage,
     35.64583333333333, 33.33333333333333, []),
    (deterministic.normalized_mean_absolute, 5.7, 5.0, [10.]),
    (deterministic.normalized_mean_bias, -0.25, 0., [10.]),
    (deterministic.normalized_root_mean_square,
     7.148776119029046, 7.071067811865476, [10.]),
]
)
def test_deadband(func, error_func_deadband, deadband_obs_fx, expect,
                  expect_deadband, args):
    obs, fx = deadband_obs_fx
    out = func(obs, fx, *args)
    out_deadband = func(obs, fx, *args, error_fnc=error_func_deadband)
    assert_allclose(out, expect)
    assert_allclose(out_deadband, expect_deadband)


@pytest.fixture()
def default_cost_err():
    # sum: 8.6, netsum: 16.6
    return pd.Series([0, 0, -1, .1, 2.5, 2.1, 1.9, -3, 6],
                     index=pd.date_range(
                         start='2020-05-01T00:00Z', freq='1h', periods=9))


@pytest.mark.parametrize('params,exp', [
    (datamodel.ConstantCost(
        cost=1.9, aggregation='sum', net=True), 8.6 * 1.9),
    (datamodel.ConstantCost(
        cost=0.11, aggregation='sum', net=False), 16.6 * 0.11),
    (datamodel.ConstantCost(
        cost=0.0, aggregation='sum', net=False), 0),
    (datamodel.ConstantCost(
        cost=2.0, aggregation='mean', net=False), 16.6/9 * 2),
    (datamodel.ConstantCost(
        cost=2.0, aggregation='mean', net=True), 8.6/9 * 2),
])
def test_constant_cost(default_cost_err, params, exp):
    fx = default_cost_err
    obs = pd.Series(0, index=fx.index)
    cost = deterministic.constant_cost(obs, fx, params,
                                       deterministic.error)
    assert cost == exp


def test_constant_cost_using_err_fnc(default_cost_err):
    params = datamodel.ConstantCost(
        cost=2.0, aggregation='sum', net=True)
    fx = default_cost_err
    obs = pd.Series(0, index=fx.index)
    cost = deterministic.constant_cost(obs, fx, params,
                                       lambda x, y: 0)
    assert cost == 0


@pytest.mark.parametrize('times,costs,index,tz,fill,exp', [
    ([dt.time(0), dt.time(2)], [1.0, 2.0],
     pd.date_range('2020-01-01T00:00Z', freq='1h', periods=3),
     'UTC', 'ffill',
     pd.Series([1.0, 1.0, 2.0])),
    # out of order times
    ([dt.time(2), dt.time(0)], [2.0, 1.0],
     pd.date_range('2020-01-01T00:00Z', freq='1h', periods=3),
     'UTC', 'ffill',
     pd.Series([1.0, 1.0, 2.0], index=pd.date_range(
         '2020-01-01T00:00Z', freq='1h', periods=3),)),
    ([dt.time(0), dt.time(4)], [1.0, 2.0],
     pd.date_range('2020-01-01T00:00Z', freq='1h', periods=3),
     'UTC', 'ffill',
     pd.Series([1.0, 1.0, 1.0])),
    ([dt.time(1), dt.time(2)], [1.0, 2.0],
     pd.date_range('2020-01-01T00:00Z', freq='1h', periods=3),
     'UTC', 'ffill',
     pd.Series([2.0, 1.0, 2.0])),
    ([dt.time(1), dt.time(1, 30)], [1.0, 2.0],
     pd.date_range('2020-01-01T00:01Z', freq='1h', periods=3),
     'UTC', 'bfill',
     pd.Series([1.0, 2.0, 1.0])),
    ([dt.time(17), dt.time(19)], [1.0, 2.0],
     pd.date_range('2020-01-01T00:00Z', freq='1h', periods=3),
     'MST', 'ffill',
     pd.Series([1.0, 1.0, 2.0])),
    # missing first time
    ([dt.time(15), dt.time(17)], [1.0, 2.0],
     pd.date_range('2020-01-01T00:00Z', freq='1h', periods=3),
     'Etc/GMT+10', 'ffill',
     pd.Series([2.0, 1.0, 1.0])),
    # missing last time
    ([dt.time(9), dt.time(11)], [1.0, 2.0],
     pd.date_range('2020-01-01T00:00Z', freq='1h', periods=3),
     'Etc/GMT-10', 'bfill',
     pd.Series([2.0, 2.0, 1.0])),
    ([dt.time(9)], [1.0],
     pd.date_range('2020-01-01T00:00Z', freq='1h', periods=2),
     'Etc/GMT-1', 'bfill',
     pd.Series([1.0, 1.0])),
    ([dt.time(0, 5), dt.time(0, 10)], [1.0, -99.9],
     pd.date_range('2020-01-01T00:00Z', freq='5min', periods=3),
     'UTC', 'ffill',
     pd.Series([-99.9, 1.0, -99.9])),
])
def test_make_time_of_day_cost_ser(times, costs, index, tz, fill, exp):
    ser = deterministic._make_time_of_day_cost_ser(times, costs, index, tz,
                                                   fill)
    if not isinstance(exp.index, pd.DatetimeIndex):
        exp.index = index
    exp.name = 'cost'
    assert_series_equal(ser, exp)


dt_cost_parametrize = pytest.mark.parametrize('fill,agg,net,tz,exp', [
    ('forward', 'sum', True, 'UTC', 16.1),
    ('backward', 'sum', True, 'UTC', 25.1),
    ('forward', 'sum', True, None, 16.1),
    ('backward', 'sum', True, None, 25.1),
    ('forward', 'mean', True, 'UTC', 16.1/9),
    ('backward', 'mean', True, 'UTC', 25.1/9),
    ('forward', 'sum', False, 'UTC', 34.1),
    ('backward', 'sum', False, 'UTC', 45.1),
    ('forward', 'mean', False, 'UTC', 34.1/9),
    ('backward', 'mean', False, 'UTC', 45.1/9),
    ('forward', 'sum', True, 'Etc/GMT+5', 13.8),
    ('backward', 'sum', True, 'Etc/GMT+5', 8.6),

])


@dt_cost_parametrize
def test_time_of_day_cost(fill, agg, net, tz, exp, default_cost_err):
    fx = default_cost_err
    obs = pd.Series(0, index=default_cost_err.index)
    params = datamodel.TimeOfDayCost(
        cost=[1.0, 2.0, 3.0],
        times=[dt.time(3), dt.time(4), dt.time(8, 1)],
        aggregation=agg,
        fill=fill,
        net=net,
        timezone=tz
    )
    out = deterministic.time_of_day_cost(obs, fx, params, deterministic.error)
    assert out == exp


def test_time_of_day_cost_alt_error_fnc(default_cost_err):
    fx = default_cost_err
    obs = pd.Series(0, index=default_cost_err.index)
    params = datamodel.TimeOfDayCost(
        cost=[1.0, 2.0, 3.0],
        times=[dt.time(3), dt.time(4), dt.time(8, 1)],
        aggregation='sum',
        fill='forward',
        net=False,
    )
    out = deterministic.time_of_day_cost(
        obs, fx, params,
        lambda x, y: pd.Series(index=pd.DatetimeIndex([]), dtype=float))
    assert out == 0


def test_time_of_day_cost_empty(default_cost_err):
    fx = default_cost_err
    obs = pd.Series(0, index=default_cost_err.index)
    params = datamodel.TimeOfDayCost(
        cost=[],
        times=[],
        aggregation='mean',
        fill='forward',
        net=True
    )
    out = deterministic.time_of_day_cost(obs, fx, params, deterministic.error)
    assert out == 0


@dt_cost_parametrize
def test_datetime_cost(fill, agg, net, tz, exp, default_cost_err):
    params = datamodel.DatetimeCost(
        datetimes=[
            pd.Timestamp('2020-04-30T12:00'),
            pd.Timestamp('2020-05-01T03:00'),
            pd.Timestamp('2020-05-01T04:00'),
            pd.Timestamp('2020-05-01T08:01'),
        ],
        cost=[3.0, 1.0, 2.0, 3.0],
        aggregation=agg,
        fill=fill,
        net=net,
        timezone=tz
    )
    fx = default_cost_err
    obs = pd.Series(0, index=default_cost_err.index)
    out = deterministic.datetime_cost(obs, fx, params, deterministic.error)
    assert out == exp


def test_datetime_cost_nans(default_cost_err):
    params = datamodel.DatetimeCost(
        datetimes=[
            pd.Timestamp('2020-05-01T03:00'),
            pd.Timestamp('2020-05-01T04:00'),
            pd.Timestamp('2020-05-01T08:01'),
        ],
        cost=[1.0, 2.0, 3.0],
        aggregation='sum',
        fill='forward',
        net=True
    )
    fx = default_cost_err
    obs = pd.Series(0, index=default_cost_err.index)
    out = deterministic.datetime_cost(obs, fx, params, deterministic.error)
    assert (out - 19.1) < 1e-9  # float precision diff

    params = datamodel.DatetimeCost(
        datetimes=[
            pd.Timestamp('2020-05-01T03:00'),
            pd.Timestamp('2020-05-01T04:00'),
        ],
        cost=[1.0, 2.0],
        aggregation='sum',
        fill='backward',
        net=True
    )
    out = deterministic.datetime_cost(obs, fx, params, deterministic.error)
    assert out == 4.1


def test_datetime_cost_alt_err(default_cost_err):
    params = datamodel.DatetimeCost(
        datetimes=[
            pd.Timestamp('2020-05-01T03:00'),
            pd.Timestamp('2020-05-01T04:00'),
            pd.Timestamp('2020-05-01T08:01'),
        ],
        cost=[1.0, 2.0, 3.0],
        aggregation='sum',
        fill='forward',
        net=True
    )
    fx = default_cost_err
    obs = pd.Series(0, index=default_cost_err.index)
    out = deterministic.datetime_cost(
        obs, fx, params,
        lambda x, y: pd.Series(index=pd.DatetimeIndex([]), dtype=float))
    assert out == 0


def test_datetime_cost_empty(default_cost_err):
    params = datamodel.DatetimeCost(
        datetimes=[],
        cost=[],
        aggregation='sum',
        fill='forward',
        net=True
    )
    fx = default_cost_err
    obs = pd.Series(0, index=default_cost_err.index)
    out = deterministic.datetime_cost(obs, fx, params,
                                      deterministic.error)
    assert out == 0


def test_band_masks(banded_cost_params, default_cost_err):
    masks = deterministic._band_masks(banded_cost_params.parameters.bands,
                                      default_cost_err)
    expmasks = [
        pd.Series([1, 1, 1, 1, 0, 0, 1, 0, 0], dtype=bool,
                  index=default_cost_err.index),
        pd.Series([0, 0, 0, 0, 1, 1, 0, 0, 1], dtype=bool,
                  index=default_cost_err.index),
        pd.Series([0, 0, 0, 0, 0, 0, 0, 1, 0], dtype=bool,
                  index=default_cost_err.index),
    ]
    assert sum(masks).all()
    for i, m in enumerate(masks):
        assert_series_equal(m, expmasks[i])


def test_error_band_cost(banded_cost_params):
    ind = pd.date_range(start='2020-05-01T00:00Z', freq='1h', periods=8)
    obs = pd.Series([1, 6, 1, 2, 2.00, 2.0, 0, 4], index=ind)
    fx = pd.Series([1., 2, 3, 0, -0.1, 1.9, 3, 5], index=ind)
    err = deterministic.error_band_cost(
        obs, fx, banded_cost_params.parameters, deterministic.error)
    exp = ((0 + 0 + 2 + -2 + 0.0 + -0.1 + 0 + 1) +
           (0 + 0 + 0 + 0. + 0.0 + 0.00 + 3 * 0.9 + 0) +
           (0 + -4 * -0.2 + 0 + 0. + -2.1 * -0.2 + 0.00 + 0 + 0))
    assert err == exp


def test_error_band_cost_positive_only(default_cost_err):
    params = datamodel.ErrorBandCost(
        bands=[datamodel.CostBand(
            error_range=(0, 10),
            cost_function='constant',
            cost_function_parameters=datamodel.ConstantCost(
                cost=2.0, aggregation='sum', net=True
            )
        )]
    )
    fx = default_cost_err
    obs = pd.Series(0, index=fx.index)
    err = deterministic.error_band_cost(
        obs, fx, params, deterministic.error)
    assert err == 25.2


def test_error_band_cost_range_equiv(default_cost_err):
    covering = datamodel.ErrorBandCost(
        bands=[
            datamodel.CostBand(
                error_range=(-2, 2),
                cost_function='constant',
                cost_function_parameters=datamodel.ConstantCost(
                    cost=2.0, aggregation='sum', net=True
                )
            ),
            datamodel.CostBand(
                error_range=(-np.inf, np.inf),
                cost_function='constant',
                cost_function_parameters=datamodel.ConstantCost(
                    cost=4.0, aggregation='sum', net=False
                )
            )
        ]
    )
    split = datamodel.ErrorBandCost(
        bands=[
            datamodel.CostBand(
                error_range=(-2, 2),
                cost_function='constant',
                cost_function_parameters=datamodel.ConstantCost(
                    cost=2.0, aggregation='sum', net=True
                )
            ),
            datamodel.CostBand(
                error_range=(-np.inf, -2),
                cost_function='constant',
                cost_function_parameters=datamodel.ConstantCost(
                    cost=4.0, aggregation='sum', net=False
                )
            ),
            datamodel.CostBand(
                error_range=(2, np.inf),
                cost_function='constant',
                cost_function_parameters=datamodel.ConstantCost(
                    cost=4.0, aggregation='sum', net=False
                )
            )
        ]
    )
    fx = default_cost_err
    obs = pd.Series(0, index=fx.index)
    err_cov = deterministic.error_band_cost(
        obs, fx, covering, deterministic.error)
    err_split = deterministic.error_band_cost(
        obs, fx, split, deterministic.error)
    assert err_cov == err_split


def test_error_band_cost_out_of_range(default_cost_err):
    params = datamodel.ErrorBandCost(
        bands=[datamodel.CostBand(
            error_range=(10, 100),
            cost_function='constant',
            cost_function_parameters=datamodel.ConstantCost(
                cost=2.0, aggregation='sum', net=True
            )
        )]
    )
    fx = default_cost_err
    obs = pd.Series(0, index=fx.index)
    err = deterministic.error_band_cost(
        obs, fx, params, deterministic.error)
    assert err == 0


def test_error_band_cost_no_bands(default_cost_err):
    params = datamodel.ErrorBandCost(
        bands=tuple()
    )
    fx = default_cost_err
    obs = pd.Series(0, index=fx.index)
    err = deterministic.error_band_cost(
        obs, fx, params, deterministic.error)
    assert err == 0


def test_error_band_cost_out_of_times(default_cost_err):
    params = datamodel.ErrorBandCost(
        bands=[datamodel.CostBand(
            error_range=(-10, 1),
            cost_function='datetime',
            cost_function_parameters=datamodel.DatetimeCost(
                datetimes=[pd.Timestamp('2020-05-01T02:00-07:00')],
                cost=[1.0],
                aggregation='sum',
                fill='forward',
                net=True
            )
        )]
    )
    fx = default_cost_err
    obs = pd.Series(0, index=fx.index)
    err = deterministic.error_band_cost(
        obs, fx, params, deterministic.error)
    assert err == 0


@pytest.mark.parametrize('params,exp', [
    ('banded', 12.06),
    (None, np.nan),
    (datamodel.Cost(
        name='cost',
        type='constant',
        parameters=datamodel.ConstantCost(
            cost=1.9, aggregation='sum', net=True)
        ), 8.6 * 1.9),
    (datamodel.Cost(
        name='cost',
        type='timeofday',
        parameters=datamodel.TimeOfDayCost(
            cost=[1.0, 2.0, 3.0],
            times=[dt.time(3), dt.time(4), dt.time(8, 1)],
            fill='forward',
            aggregation='sum',
            net=True
        )), 16.1),
    (datamodel.Cost(
        name='cost',
        type='datetime',
        parameters=datamodel.DatetimeCost(
            datetimes=[
                pd.Timestamp('2020-04-30T12:00'),
                pd.Timestamp('2020-05-01T03:00'),
                pd.Timestamp('2020-05-01T04:00'),
                pd.Timestamp('2020-05-01T08:01'),
            ],
            cost=[3.0, 1.0, 2.0, 3.0],
            aggregation='sum',
            fill='backward',
            net=True,
            timezone='Etc/GMT+5'
        )), 8.6)
])
def test_cost(params, exp, banded_cost_params, default_cost_err):
    fx = default_cost_err
    obs = pd.Series(0, index=fx.index)

    if params == 'banded':
        params = banded_cost_params

    out = deterministic.cost(obs, fx, params)
    if np.isnan(exp):
        assert np.isnan(out)
    else:
        assert out == exp



"""Collection of code for requesting and parsing Data from the DOE
RTC pv-dashboard hosted by Sandia National Laboratories
"""
import logging


import pandas as pd
import requests

DOE_RTC_API_URL = "https://pv-dashboard.sandia.gov/api/v1.0/location/{location}/data/{data_type}/start/{start}/end/{end}/key/{api_key}" # NOQA
logger = logging.getLogger(__name__)


def request_doe_rtc_data(location, data_type, start, end, api_key):
    """Makes a request to DOE RTC pv dashboard with the provided parameters.

    Parameters
    ----------
    location: string
        Name of the DOE RTC location.
    data_type: string
        'system' or 'weather'
    api_key: string
        The Api key for accessing the RTC pv dashboard API.
    start: datetime
        Beginning of the period for which to request data.
    end: datetime
        End of the period for which to request data.

    Returns
    -------
    DataFrame
        DataFrame parsed from the json response.
    """
    request_url = DOE_RTC_API_URL.format(
        location=location,
        start=start.strftime('%Y-%m-%d'),
        end=end.strftime('%Y-%m-%d'),
        data_type=data_type,
        api_key=api_key)
    r = requests.get(request_url)
    resp = r.json()
    if 'access' in resp and resp['access'] == 'denied':
        raise ValueError('Invalid DOE RTC API key')
    else:
        return pd.DataFrame(resp)


def fetch_doe_rtc(location, api_key, start, end):
    """
    Requests and concatenates data from the DOE RTC pv dashboard API
    into a single dataframe.

    Parameters
    ----------
    location: string
        Name of the DOE RTC location.
    api_key: string
        The Api key for accessing the DOE RTC API.
    start: datetime
        Beginning of the period for which to request data.
    end: datetime
        End of the period for which to request data.

    Returns
    -------
    pandas.DataFrame
        With data from start to end. Index is a datetime-index NOT localized
        but in the timezone for the location
    """
    data_types = ['system', 'weather']
    dfs = []
    for data_type in data_types:
        df = request_doe_rtc_data(location, data_type, start, end, api_key)
        if df.empty:
            continue
        df.index = pd.to_datetime(df['TmStamp'], unit='ms', utc=False)
        df = df.drop('TmStamp', axis=1)
        # Append the datatype to the end of AmbientTemp so we can differentiate
        # system from weather temperatures
        df = df.rename(columns={'AmbientTemp': f'AmbientTemp_{data_type}'})
        dfs.append(df)
    try:
        data = pd.concat(dfs, axis=1)
    except ValueError:
        # empty data
        return pd.DataFrame()
    return data



"""
Functions to make all of the metrics figures for Solar Forecast Arbiter reports
using Plotly.
"""
import base64
import calendar
from copy import deepcopy
import datetime as dt
from itertools import cycle
from pathlib import Path
import logging


import pandas as pd
from plotly import __version__ as plotly_version
import plotly.graph_objects as go
import numpy as np
from matplotlib import cm
from matplotlib.colors import Normalize


from solarforecastarbiter import datamodel
from solarforecastarbiter.metrics.event import _event2count
import solarforecastarbiter.plotting.utils as plot_utils


logger = logging.getLogger(__name__)
D3_PALETTE = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b',
              '#e377c2', '#7f7f7f', '#bcbd22', '#17becf', '#aec7e8', '#ffbb78',
              '#98df8a', '#ff9896', '#c5b0d5', '#c49c94', '#f7b6d2', '#c7c7c7',
              '#dbdb8d', '#9edae5']
PALETTE = (D3_PALETTE[::2] + D3_PALETTE[1::2])


def gen_grays(num_colors):
    """Generate a grayscale color list of length num_colors.
    """
    rgb_delta = int(255/num_colors + 1)
    color_list = ["#{h}{h}{h}".format(h=hex(i*rgb_delta)[2:])
                  for i in range(num_colors)]
    return color_list


_num_obs_colors = 3
# drop white
OBS_PALETTE = gen_grays(_num_obs_colors)
OBS_PALETTE.reverse()
OBS_PALETTE_TD_RANGE = pd.timedelta_range(
            freq='10min', end='60min', periods=_num_obs_colors)

# list of matplotlib's perceptually uniform sequential color pallettes
PROBABILISTIC_PALETTES = ['viridis', 'plasma', 'inferno', 'magma', 'cividis']

PLOT_BGCOLOR = '#FFF'
PLOT_MARGINS = {'l': 50, 'r': 50, 'b': 50, 't': 50, 'pad': 4}
PLOT_LAYOUT_DEFAULTS = {
    'autosize': True,
    'height': 250,
    'margin': PLOT_MARGINS,
    'plot_bgcolor': PLOT_BGCOLOR,
    'font': {'size': 14}
}

# If for some reason, the fail.pdf (just a pdf with some text that
# pdf generation failed) is unavailable, use an empty pdf
try:
    with open(Path(__file__).parent / 'fail.pdf', 'rb') as f:
        fail_pdf = base64.a85encode(f.read()).decode()
except Exception:
    fail_pdf = ',u@!!/MSk8$73+IY58P_+>=pV@VQ644<Q:NASu.&BHT/T0Ha7#+<Vd[7VQ[\\ATAnH7VlLTAOL*>De*Dd5!B<pFE1r$D$kNX1K6%.6<uqiV.X\\GOIXKoa;)c"!&3^A=pehYA92j5ARTE_ASu$s@VQ6-+>=pV@VQ5m+<WEu$>"*cDdmGg1E\\@oDdmGg4?Ns74pkk=A8bpl$8N_X+E(_($9UEn03!49AKWX&@:s-o,p4oL+<Vd[:gnBUDKI!U+>=p9$6UH6026"gBjj>HGT^350H`%l0J5:A+>>E,2\'?03+<Vd[6Z6jaASuU2+>b2p+ArOh+<W=-Ec6)>+?Van+<VdL+<W=:H#R=;01U&$F`7[1+<VdL+>6Y902ut#DKBc*Eb0,uGmYZ:+<VdL01d:.Eckq#+<VdL+<W=);]m_]AThctAPu#b$6UH65!B;r+<W=8ATMd4Ear[%+>Y,o+ArP14pkk=A8bpl$8EYW+E(_($9UEn03!49AKWX&@:s.m$6UH602$"iF!+[01*A7n;BT6P+<Vd[6Z7*bF<E:F5!B<bDIdZpC\'ljA0Hb:CC\'m\'c+>6Q3De+!#ATAnA@ps(lD]gbe0fCX<+=LoFFDu:^0/$gDBl\\-)Ea`p#Bk)3:DfTJ>.1.1?+>6*&ART[pDf.sOFCcRC6om(W1,(C>0K1^?0ebFC/MK+20JFp_5!B<bDIdZpC\'lmB0Hb:CC\'m\'c+>6]>E+L.F6Xb(FCi<qn+<Vd[:gn!JF!*1[0Ha7#5!B<bDIdZpC\'o3+AS)9\'+?0]^0JG170JG170H`822)@*4AfqF70JG170JG:B0d&/(0JG1\'DBK9?0JG170JG4>0d&/(0JG1\'DBK9?0JG170JG4<0H`&\'0JG1\'DBK9?0JG170JG182\'=S,0JG1\'DBK9?0JG170JG493?U"00JG1\'DBK9?0JG170JG=?2BX\\-0JG1\'DBK9?0JG170JG@B1*A8)0JG1\'DBK:.Ea`ZuATA,?4<Q:UBmO>53!pcN+>6W2Dfd*\\+>=p9$6UH601g%nD]gq\\0Ha7#5!B<pFCB33G]IA-$8sUq$7-ue:IYZ'  # NOQA


def _value_frame_dict(idx, pfxobs, column=None):
    if column is None:
        forecast_values = pfxobs.forecast_values
    else:
        forecast_values = pfxobs.forecast_values[column]
    value_frame_dict = {
        'pair_index': idx,
        'observation_values': pfxobs.observation_values,
        'forecast_values': forecast_values,
    }
    return value_frame_dict


def _meta_row_dict(idx, pfxobs, **kwargs):
    forecast_object = kwargs.pop('forecast_object', None)
    if forecast_object is None:
        forecast_object = pfxobs.original.forecast

    # Check for a case where we're adding metadata for a constant value, but
    # the pair contains a whole ProbabilisticForecast
    if (isinstance(forecast_object,
                   datamodel.ProbabilisticForecastConstantValue)
        and
        isinstance(pfxobs.original.forecast,
                   datamodel.ProbabilisticForecast)):
        distribution = str(hash((
            pfxobs.original.forecast,
            pfxobs.original.forecast.interval_length,
            pfxobs.original.forecast.interval_value_type,
            pfxobs.original.forecast.interval_label)))
    else:
        distribution = None
    try:
        axis = forecast_object.axis
    except AttributeError:
        axis = None
    try:
        constant_value = forecast_object.constant_value
    except AttributeError:
        constant_value = None
    meta = {
        'pair_index': idx,
        'observation_name': _obs_name(pfxobs.original),
        'forecast_name': _fx_name(
            forecast_object, pfxobs.original.data_object),
        'interval_label': pfxobs.interval_label,
        'interval_length': pfxobs.interval_length,
        'forecast_type': pfxobs.original.__class__.__name__,
        'axis': axis,
        'constant_value': constant_value,
        'observation_hash': str(hash((
            pfxobs.original.data_object,
            pfxobs.interval_length,
            pfxobs.interval_value_type,
            pfxobs.interval_label))),
        'forecast_hash': str(hash((
            forecast_object,
            pfxobs.interval_length,
            pfxobs.interval_value_type,
            pfxobs.interval_label))),
        'observation_color': _obs_color(
            pfxobs.interval_length),
        'distribution': distribution
    }
    meta.update(kwargs)
    return meta


def construct_timeseries_dataframe(report):
    """Construct two standardized Dataframes for the timeseries and scatter
    plot functions. One with timeseries data for all observations,
    aggregates, and forecasts in the report, and the other with
    associated metadata sharing a common `pair_index` key.

    Parameters
    ----------
    report: :py:class:`solarforecastarbiter.datamodel.Report`

    Returns
    -------
    data : pandas.DataFrame
        Keys are an integer `pair_index` for pairing values with the metadata
        in the metadata_cds, and two pandas.Series, `observation_values` and
        `forecast_values`.

    metadata : pandas.DataFrame
        This dataframe has the following columns:

        - `pair_index`: Integer for pairing metadata with the values in the data dataframe.
        - `observation_name`: Observation name.
        - `forecast_name`: Forecast name.
        - `interval_label`: Interval label of the processed forecast and observation data.
        - `observation_hash`: Hash of the original observation object and the `datamodel.ProcessedForecastObservations` metadata.
        - `forecast_hash`: Hash of the original forecast object and the `datamodel.ProcessedForecastObservations` metadata.

    """  # NOQA
    value_frames = []
    meta_rows = []
    # enumerate won't work because of the conditional for loop, so
    # manually keep track of the index
    idx = 0
    for pfxobs in report.raw_report.processed_forecasts_observations:
        if isinstance(pfxobs.original.forecast,
                      datamodel.ProbabilisticForecast):
            for cvfx in pfxobs.original.forecast.constant_values:
                value_frame_dict = _value_frame_dict(
                    idx, pfxobs, column=str(cvfx.constant_value))
                # specify fx type so we know the const value fx came from a
                # ProbabilisticForecast
                meta_row_dict = _meta_row_dict(
                    idx, pfxobs,
                    forecast_object=cvfx,
                    forecast_type='ProbabilisticForecast')
                value_frames.append(pd.DataFrame(value_frame_dict))
                meta_rows.append(meta_row_dict)
                idx += 1
        else:
            value_frame_dict = _value_frame_dict(idx, pfxobs)
            meta_row_dict = _meta_row_dict(idx, pfxobs)
            value_frames.append(pd.DataFrame(value_frame_dict))
            meta_rows.append(meta_row_dict)
            idx += 1
    data = pd.concat(value_frames)
    metadata = pd.DataFrame(meta_rows)
    # convert data to report timezone
    data = data.tz_convert(report.raw_report.timezone)
    data = data.rename_axis('timestamp')
    return data, metadata


def _fill_timeseries(df, interval_length):
    """Returns a dataframe with a datetimeindex with regular frequency of
    interval_length minutes. Previously missing values will be filled with
    nans. Useful for creating gaps in plotted timeseries data.

    Parameters
    ----------
    df: pandas.DataFrame
        Dataframe with timeseries data.
    interval_length: numpy.timedelta64
        Interval length of the processed forecast observation.

    Returns
    -------
    pandas.DataFrame
        DataFrame with filled datetime index data.
    """
    if not df.index.empty:
        start = df.index[0]
        end = df.index[-1]
        freq_mins = int(interval_length / np.timedelta64(1, 'm'))
        filled_idx = pd.date_range(start, end, freq=f'{freq_mins}min')
        return df.reindex(filled_idx)
    else:
        return df


def _obs_name(fx_obs):
    # TODO: add code to ensure obs names are unique
    name = fx_obs.data_object.name
    if fx_obs.forecast.name == fx_obs.data_object.name:
        if isinstance(fx_obs.data_object, datamodel.Observation):
            name += ' Observation'
        else:
            name += ' Aggregate'
    return name


def _fx_name(forecast, data_object):
    # TODO: add code to ensure fx names are unique
    forecast_name = forecast.name
    if isinstance(forecast, datamodel.ProbabilisticForecastConstantValue):
        if forecast.axis == 'x':
            forecast_name += \
                f' Prob(x <= {forecast.constant_value} {forecast.units})'
        else:
            forecast_name += f' Prob(f <= x) = {forecast.constant_value}%'
    if forecast_name == data_object.name:
        forecast_name += ' Forecast'
    return forecast_name


def _obs_color(interval_length):
    idx = np.searchsorted(OBS_PALETTE_TD_RANGE, interval_length)
    obs_color = OBS_PALETTE[idx]
    return obs_color


def _boolean_filter_indices_by_pair(value_cds, pair_index):
    return value_cds.data['pair_index'] == pair_index


def _none_or_values0(metadata, key):
    value = metadata.get(key)
    if value is not None:
        value = value.values[0]
    return value


def _extract_metadata_from_df(metadata_df, hash_, hash_key):
    metadata = metadata_df[metadata_df[hash_key] == hash_]
    meta = {
        'pair_index': metadata['pair_index'].values[0],
        'observation_name': metadata['observation_name'].values[0],
        'forecast_name': metadata['forecast_name'].values[0],
        'interval_label': metadata['interval_label'].values[0],
        'interval_length': metadata['interval_length'].values[0],
        'observation_color': metadata['observation_color'].values[0],
    }
    meta['forecast_type'] = _none_or_values0(metadata, 'forecast_type')
    meta['axis'] = _none_or_values0(metadata, 'axis')
    meta['constant_value'] = _none_or_values0(metadata, 'constant_value')
    return meta


def _legend_text(name, max_length=20):
    """Inserts <br> tags in a name to mimic word-wrap behavior for long names
    in the legend of timeseries plots.

    Parameters
    ----------
    name: str
        The name/string to apply word-wrap effect to.
    max_length: int
        The maximum length of any line of text. Note that this will not break
        words across lines, but on the closest following space.

    Returns
    -------
    str
        The name after it is split appropriately.
    """
    if len(name) > max_length:
        temp = []
        new = []
        for part in name.split(' '):
            if len(' '.join(temp + [part])) > max_length:
                new.append(' '.join(temp))
                temp = [part]
            else:
                temp.append(part)
        if temp:
            new.append(' '.join(temp))
        return '<br>'.join(new)
    else:
        return name


def _plot_obs_timeseries(fig, timeseries_value_df, timeseries_meta_df):
    # construct graph objects in random hash order. collect them in a list
    # along with the pair index. Then add traces in order of pair index.
    gos = []
    # construct graph objects in random hash order
    for obs_hash in np.unique(timeseries_meta_df['observation_hash']):
        metadata = _extract_metadata_from_df(
            timeseries_meta_df, obs_hash, 'observation_hash')
        pair_idcs = timeseries_value_df['pair_index'] == metadata['pair_index']
        plot_kwargs = plot_utils.line_or_step_plotly(
            metadata['interval_label'])
        data = _fill_timeseries(
            timeseries_value_df[pair_idcs],
            metadata['interval_length'],
        )
        go_ = go.Scattergl(
            y=data['observation_values'],
            x=data.index,
            name=_legend_text(metadata['observation_name']),
            legendgroup=metadata['observation_name'],
            marker=dict(color=metadata['observation_color']),
            connectgaps=False,
            **plot_kwargs)
        # collect in list
        gos.append((metadata['pair_index'], go_))
    # Add traces in order of pair index
    for idx, go_ in sorted(gos, key=lambda x: x[0]):
        fig.add_trace(go_)


def _plot_fx_timeseries(fig, timeseries_value_df, timeseries_meta_df, axis):
    palette = cycle(PALETTE)
    # pull metadata to plot in random hash order. collect them in a list
    # along with the pair index. Then add traces in order of pair index.
    metadatas = []

    # pull metadata to plot in random hash order
    for fx_hash in np.unique(timeseries_meta_df['forecast_hash']):
        metadata = _extract_metadata_from_df(
            timeseries_meta_df, fx_hash, 'forecast_hash')
        if metadata['axis'] not in axis:
            # we're looking at a different kind of forecast than what we wanted
            # to plot
            continue
        # collect in list
        metadatas.append((metadata['pair_index'], metadata))

    for idx, metadata in sorted(metadatas, key=lambda x: x[0]):
        pair_idcs = timeseries_value_df['pair_index'] == metadata['pair_index']
        # probably treat axis == None and axis == y separately in the future.
        # currently no need for a separate axis == x treatment either, so
        # removed an if statement on the axis.
        plot_kwargs = plot_utils.line_or_step_plotly(
            metadata['interval_label'])
        data = _fill_timeseries(
            timeseries_value_df[pair_idcs],
            metadata['interval_length'],
        )
        plot_kwargs['marker'] = dict(color=next(palette))
        go_ = go.Scattergl(
            y=data['forecast_values'],
            x=data.index,
            name=_legend_text(metadata['forecast_name']),
            legendgroup=metadata['forecast_name'],
            connectgaps=False,
            **plot_kwargs)
        fig.add_trace(go_)


def _plot_fx_distribution_timeseries(
        fig, timeseries_value_df, timeseries_meta_df, axis):
    palette = cycle(PROBABILISTIC_PALETTES)
    gos = []

    for dist_hash in np.unique(timeseries_meta_df['distribution']):
        # indices to constant values in the metadata df
        cv_indices = timeseries_meta_df['distribution'] == dist_hash

        # sort constant values
        cv_metadata = timeseries_meta_df[cv_indices]
        cv_metadata = cv_metadata.sort_values('constant_value')
        cv_metadata = cv_metadata.reset_index()

        # Get a colormap for mapping fill colors
        color_map = cm.get_cmap(next(palette))
        color_scaler = cm.ScalarMappable(
            Normalize(vmin=0, vmax=1),
            color_map,
        )

        symmetric_percentiles = plot_utils.percentiles_are_symmetric(
            cv_metadata['constant_value'].tolist())
        # Plot confidence intervals
        for idx, cv in cv_metadata.iterrows():
            pair_idcs = timeseries_value_df['pair_index'] == cv['pair_index']
            data = _fill_timeseries(
                timeseries_value_df[pair_idcs],
                cv['interval_length'])

            # Fill missing data with 0 to avoid plotly bugs encountered with
            # go.Scatter fill and missing data.
            data = data.fillna(0)

            if idx == 0:
                # The first value will act as the lower bound for other values
                # to fill down to.
                fill = None
                showlegend = True
            else:
                fill = 'tonexty'
                showlegend = False

            # Split name of the distribution from the current constant value
            constant_label_index = cv['forecast_name'].find('Prob(') - 1
            fx_name = cv['forecast_name'][:constant_label_index]
            cv_label = cv['forecast_name'][constant_label_index:]

            if symmetric_percentiles:
                # Since plotly always fills below the line, for constants below
                # 50%, use the previous value to mimic fill upward behavior.
                # E.g. fill downward from 5% to 0% with the 100% interval.
                if cv['constant_value'] <= 50 and idx != 0:
                    fill_value = cv_metadata.iloc[idx - 1]['constant_value']
                else:
                    fill_value = cv['constant_value']

                # When constant values are symmetric, create intervals
                # centered around the 50th percentile
                fill_value = 2 * abs(fill_value - 50)
            else:
                # convert to complement percentile to invert shading, such that
                # bright colors appear at 0 and dark at 100 when plotted.
                fill_value = 100 - cv['constant_value']

            fill_color = plot_utils.distribution_fill_color(
                color_scaler, fill_value)

            plot_kwargs = plot_utils.line_or_step_plotly(cv['interval_label'])

            go_ = go.Scatter(
                x=data.index,
                y=data['forecast_values'],
                name=_legend_text(fx_name),
                hovertemplate=(
                    f'<b>{ cv_label }<br>'
                    '<b>Value<b>: %{y}<br>'
                    '<b>Time<b>: %{x}<br>'),
                connectgaps=False,
                mode='lines',
                fill=fill,
                showlegend=showlegend,
                legendgroup=cv['distribution'],
                fillcolor=fill_color,
                line=dict(
                    color=fill_color,
                ),
                **plot_kwargs,
            )

            # Add traces in order of pair index
            gos.append((cv['pair_index'], go_))
    for idx, go_ in sorted(gos, key=lambda x: x[0]):
        fig.add_trace(go_)


def timeseries(timeseries_value_df, timeseries_meta_df,
               start, end, units, axis, timezone='UTC'):
    """
    Timeseries plot of one or more forecasts and observations.

    Parameters
    ----------
    timeseries_value_df: pandas.DataFrame
        DataFrame of timeseries data. See
        :py:func:`solarforecastarbiter.reports.figures.construct_timeseries_dataframe`
        for format.
    timeseries_meta_df: pandas.DataFrame
        DataFrame of metadata for each Observation Forecast pair. See
        :py:func:`solarforecastarbiter.reports.figures.construct_timeseries_dataframe`
        for format.
    start : pandas.Timestamp
        Report start time
    end : pandas.Timestamp
        Report end time
    axis : {(None,), ('x',), ('y',), (None, 'y')}
        Specifies the kinds of forecast to plot. None is appropriate for
        deterministic forecasts, 'x' for probabilistic forecasts with
        axis = 'x', and 'y' for probabilistic forecasts with
        axis = 'y'. Observations, deterministic forecasts, and
        probabilistic forecasts may all be plotted together if
        axis = (None, 'y'). Observations will not be plotted if
        axis = ('x',).
    timezone : str
        Timezone consistent with the data in the timeseries_metadata_df.

    Returns
    -------
    plotly.Figure
    """  # NOQA: E501
    # might want to make fig=None a kwarg and modify this line to
    # fig = fig if fig is not None else go.Figure()
    fig = go.Figure()

    if 'x' in axis:
        ylabel = 'Probability (%)'
    else:
        ylabel = f'Data ({units})'
        # adds observation traces to fig
        _plot_obs_timeseries(fig, timeseries_value_df, timeseries_meta_df)

    # add forecast traces that have correct axis to fig
    # get indices of probabilistic forecasts with axis y to create special
    # shaded distribution plots
    y_distribution_indices = (
        timeseries_meta_df['distribution'].notna()
        & (timeseries_meta_df['axis'] == 'y')
    )
    non_y_distribution_meta_df = timeseries_meta_df[~y_distribution_indices]
    distribution_meta_df = timeseries_meta_df[y_distribution_indices]

    _plot_fx_timeseries(
        fig, timeseries_value_df, non_y_distribution_meta_df, axis)
    _plot_fx_distribution_timeseries(
        fig, timeseries_value_df, distribution_meta_df, axis)

    fig.update_xaxes(title_text=f'Time ({timezone})', showgrid=True,
                     gridwidth=1, gridcolor='#CCC', showline=True,
                     linewidth=1, linecolor='black', ticks='outside')
    fig.update_yaxes(title_text=ylabel, showgrid=True,
                     gridwidth=1, gridcolor='#CCC', showline=True,
                     linewidth=1, linecolor='black', ticks='outside',
                     fixedrange=True)
    fig.update_layout(
        legend=dict(font=dict(size=10)),
    )
    return fig


def _get_scatter_limits(df):
    extremes = [np.nan]
    for kind in ('forecast_values', 'observation_values'):
        arr = np.asarray(df[kind]).astype(float)
        if len(arr) != 0:
            extremes.append(np.nanmin(arr))
            extremes.append(np.nanmax(arr))
    min_ = np.nanmin(extremes)
    if np.isnan(min_):
        min_ = -999
    max_ = np.nanmax(extremes)
    if np.isnan(max_):
        max_ = 999
    return min_, max_


def scatter(timeseries_value_df, timeseries_meta_df, units):
    """
    Adds Scatter plot traces of one or more forecasts and observations to
    the figure.

    Parameters
    ----------
    timeseries_value_df: pandas.DataFrame
        DataFrame of timeseries data. See
        :py:func:`solarforecastarbiter.reports.figures.construct_timeseries_dataframe`
        for format.
    timeseries_meta_df: pandas.DataFrame
        DataFrame of metadata for each Observation Forecast pair. See
        :py:func:`solarforecastarbiter.reports.figures.construct_timeseries_dataframe`
        for format.

    Returns
    -------
    plotly.Figure
    """  # NOQA
    scatter_range = _get_scatter_limits(timeseries_value_df)

    palette = cycle(PALETTE)
    fig = go.Figure()
    # pull metadata to plot in random hash order. collect them in a list
    # along with the pair index. Then add traces in order of pair index.
    metadatas = []
    # accumulate labels and plot objects for manual legend
    for fxhash in np.unique(timeseries_meta_df['forecast_hash']):
        metadata = _extract_metadata_from_df(
            timeseries_meta_df, fxhash, 'forecast_hash')
        if metadata['axis'] == 'x':
            # don't know how to represent probability forecasts on a
            # physical value vs. physical value plot.
            continue
        # collect in list
        metadatas.append((metadata['pair_index'], metadata))

    # plot in order of pair index
    for idx, metadata in sorted(metadatas, key=lambda x: x[0]):
        pair_idcs = timeseries_value_df['pair_index'] == metadata['pair_index']
        data = timeseries_value_df[pair_idcs]
        go_ = go.Scattergl(
            x=data['observation_values'],
            y=data['forecast_values'],
            name=_legend_text(metadata['forecast_name']),
            showlegend=True,
            legendgroup=metadata['forecast_name'],
            marker=dict(color=next(palette), opacity=0.25),
            mode='markers')
        fig.add_trace(go_)

    label = f'({units})'
    x_label = 'Observed ' + label
    y_label = 'Forecast ' + label
    nticks = 10
    fig.update_xaxes(title_text=x_label, showgrid=True,
                     gridwidth=1, gridcolor='#CCC', showline=True,
                     linewidth=1, linecolor='black', ticks='outside',
                     range=scatter_range, nticks=nticks)
    fig.update_yaxes(title_text=y_label, showgrid=True,
                     gridwidth=1, gridcolor='#CCC', showline=True,
                     linewidth=1, linecolor='black', ticks='outside',
                     range=scatter_range, nticks=nticks)
    return fig


def event_histogram(timeseries_value_df, timeseries_meta_df):
    """
    Adds histogram plot traces of the event outcomes of one or more event
    forecasts and observations to the figure.

    Parameters
    ----------
    timeseries_value_df: pandas.DataFrame
        DataFrame of timeseries data. See
        :py:func:`solarforecastarbiter.reports.figures.construct_timeseries_dataframe`
        for format.
    timeseries_meta_df: pandas.DataFrame
        DataFrame of metadata for each Observation Forecast pair. See
        :py:func:`solarforecastarbiter.reports.figures.construct_timeseries_dataframe`
        for format.

    Returns
    -------
    plotly.Figure
    """  # NOQA

    fig = go.Figure()
    palette = cycle(PALETTE)
    # accumulate labels and plot objects for manual legend
    for fxhash in np.unique(timeseries_meta_df['forecast_hash']):
        metadata = _extract_metadata_from_df(
            timeseries_meta_df, fxhash, 'forecast_hash')
        pair_idcs = timeseries_value_df['pair_index'] == metadata['pair_index']
        data = timeseries_value_df[pair_idcs]

        tp, fp, tn, fn = _event2count(data["observation_values"],
                                      data["forecast_values"])
        x = ["True Pos.", "False Pos.", "True Neg.", "False Neg."]
        y = [tp, fp, tn, fn]

        fig.add_trace(go.Bar(
            x=x,
            y=y,
            name=_legend_text(metadata['forecast_name']),
            showlegend=True,
            legendgroup=metadata['forecast_name'],
            marker_color=next(palette),
        ))

    # update axes
    x_label = "Outcome"
    y_label = "Count"
    fig.update_xaxes(title_text=x_label, showgrid=True,
                     gridwidth=0, gridcolor='#CCC', showline=True,
                     linewidth=1, linecolor='black', ticks='outside')
    fig.update_yaxes(title_text=y_label, showgrid=True,
                     gridwidth=1, gridcolor='#CCC', showline=True,
                     linewidth=1, linecolor='black', ticks='outside')

    return fig


def configure_axes(fig, x_axis_kwargs, y_axis_kwargs):
    """Applies plotly axes configuration to display zero line and grid, and the
    configuration passed in x_axis_kwargs and y_axis kwargs. Currently
    configured to supply base layout for metric plots.

    Parameters
    ----------
    fig: plotly.graph_objects.Figure

    x_axis_kwargs: dict
        Dictionary to expand as arguments to fig.update_xaxes.
    y_axis_kwargs: dict
        Dictionary to expand as arguments to fig.update_x_axes.
    """
    fig.update_xaxes(showline=True, linewidth=1, linecolor='black',
                     ticks='outside')
    if x_axis_kwargs:
        fig.update_xaxes(**x_axis_kwargs)
    fig.update_yaxes(showgrid=True, gridwidth=1, gridcolor='#CCC')
    fig.update_yaxes(showline=True, linewidth=1, linecolor='black',
                     ticks='outside')
    if y_axis_kwargs:
        fig.update_yaxes(**y_axis_kwargs)


def construct_metrics_dataframe(metrics, rename=None):
    """
    Possibly bad assumptions:
    * metrics contains keys: name, Total, etc.

    Parameters
    ----------
    metrics : list of datamodel.MetricResults
        Each metric dict is for a different forecast. Forecast name is
        specified by the name key.
    rename : function or None
        Function of one argument that is applied to each forecast name.

    Returns
    -------
    df: pandas.DataFrame
        Dataframe of computed metrics for the report.
    """

    if rename:
        f = rename
    else:
        def f(x): return x  # NOQA

    data = []
    for metric_result in metrics:
        for mvalue in metric_result.values:
            new = {
                'name': metric_result.name,
                'abbrev': f(metric_result.name),
                'category': mvalue.category,
                'metric': mvalue.metric,
                'value': mvalue.value
            }
            if new['category'] == 'date':
                new['index'] = dt.datetime.strptime(
                    mvalue.index, '%Y-%m-%d')
            else:
                new['index'] = mvalue.index
            data.append(new)
    df = pd.DataFrame(data, columns=[
        'name', 'abbrev', 'category', 'metric', 'value', 'index'
    ])
    return df


def abbreviate(x, limit=3):
    # might need to add logic to ensure uniqueness
    # and/or enforce max length using textwrap.shorten
    components = x.split(' ')
    out_components = []
    for c in components:
        if len(c) <= limit:
            out = c
        elif c.upper() == c:
            # probably an acronym
            out = c
        else:
            out = f'{c[0:limit]}.'
        out_components.append(out)
    return ' '.join(out_components)


def bar(df, metric):
    """
    Create a bar graph comparing a single metric across forecasts.

    Parameters
    ----------
    df: pandas Dataframe
        Metric dataframe by :py:func:`solarforecastarbiter.reports.figures.construct_metrics_dataframe`
    metric: str
        The metric to plot. This value should be found in df['metric'].

    Returns
    -------
    plotly.Figure
        A bar chart representing the total category of the metric for each
        forecast.
    """  # NOQA
    data = df[(df['category'] == 'total') & (df['metric'] == metric)]
    y_range = None
    x_axis_kwargs = {}
    x_values = data['abbrev']
    palette = cycle(PALETTE)
    palette = [next(palette) for _ in x_values]
    metric_name = datamodel.ALLOWED_METRICS[metric]

    # remove height limit when long abbreviations are used or there are more
    # than 5 pairs to problems with labels being cutt off.
    plot_layout_args = deepcopy(PLOT_LAYOUT_DEFAULTS)
    if x_values.map(len).max() > 15 or x_values.size > 6:
        # Set explicit height and set automargin on x axis to allow for dynamic
        # sizing to accomodate long x axis labels
        plot_layout_args['height'] = 600
        x_axis_kwargs = {'automargin': True}
    fig = go.Figure()
    fig.add_trace(go.Bar(x=x_values, y=data['value'],
                         marker=go.bar.Marker(color=palette)))
    fig.update_layout(
        title=f'<b>{metric_name}</b>',
        xaxis_title=metric_name,
        **plot_layout_args)
    configure_axes(fig, x_axis_kwargs, y_range)
    return fig


def calc_y_start_end(y_min, y_max, pad_factor=1.03):
    """
    Determine y axis start, end.

    Parameters
    ----------
    y_min : float
    y_max : float
    pad_factor : float
        Number by which to multiply the start, end.

    Returns
    -------
    start, end : float, float
    """
    # limits cannot be nans or infs
    y_min = np.nan_to_num(y_min)
    y_max = np.nan_to_num(y_max)

    if y_max < 0:
        # all negative, so set range from y_min to 0
        start = y_min
        end = 0
    elif y_min > 0:
        # all positive, so set range from 0 to y_max
        start = 0
        end = y_max
    else:
        start = y_min
        end = y_max

    # if y_max or min was +/- inf then padding will result in overflow
    # that can be ignored
    with np.errstate(over='ignore'):
        start, end = pad_factor * start, pad_factor * end

    return start, end


def bar_subdivisions(df, category, metric):
    """
    Create bar graphs comparing a single metric across subdivisions of
    time for multiple forecasts. e.g.::

        Fx 1 MAE |
                 |_________________
        Fx 2 MAE |
                 |_________________
                   Year, Month of the year, etc.

    Parameters
    ----------
    df: pandas.DataFrame
        Fields must be kind and the names of the forecasts
    category : str
        One of the available metrics grouping categories (e.g., total)
    metric : str
        One of the available metrics (e.g. mae)

    Returns
    -------
    figs : dict of figures
    """
    palette = cycle(PALETTE)

    figs = {}

    human_category = datamodel.ALLOWED_CATEGORIES[category]
    metric_name = datamodel.ALLOWED_METRICS[metric]

    x_axis_label = human_category
    y_axis_label = metric_name

    data = df[(df['category'] == category) & (df['metric'] == metric)]

    x_offset = None

    # Special handling for x-axis with dates
    if category == 'weekday':
        x_ticks = calendar.day_abbr[0:]
        x_axis_kwargs = {'tickvals': x_ticks,
                         'range': (-.5, len(x_ticks))}
    elif category == 'hour':
        x_ticks = list(range(25))
        x_axis_kwargs = {'tickvals': x_ticks,
                         'range': (-.5, len(x_ticks))}
        # plotly's offset of 0, makes the bars left justified at the tick
        x_offset = 0
    elif category == 'year':
        x_axis_kwargs = {'dtick': 1}
    elif category == 'date':
        x_axis_kwargs = {'dtick': '1d'}
    else:
        x_axis_kwargs = {}

    y_data = np.asarray(data['value'])
    if len(y_data) == 0 or np.isnan(y_data).all():
        y_range = (None, None)
    else:
        y_min = np.nanmin(y_data)
        y_max = np.nanmax(y_data)
        y_range = calc_y_start_end(y_min, y_max)
    y_axis_kwargs = {'range': y_range}
    unique_names = np.unique(np.asarray(data['name']))
    palette = [next(palette) for _ in unique_names]
    for i, name in enumerate(unique_names):
        plot_data = data[data['name'] == name]
        if len(plot_data['index']):
            x_values = plot_data['index']
        else:
            x_values = []
        if category == 'weekday':
            # Fill with mon-fri values and pass to enforce displaying the full
            # week of data.
            y_values = [plot_data[plot_data['index'] == day]['value'].iloc[0]
                        if not plot_data[plot_data['index'] == day].empty
                        else np.nan for day in x_ticks]
            x_values = x_ticks
        else:
            y_values = plot_data['value']
        # Create figure
        title = name + ' ' + metric_name
        fig = go.Figure()
        fig.add_trace(go.Bar(x=x_values, y=y_values, offset=x_offset,
                             marker=go.bar.Marker(color=palette[i])))

        fig.update_layout(
            title=f'<b>{title}</b>',
            xaxis_title=x_axis_label,
            yaxis_title=y_axis_label,
            **PLOT_LAYOUT_DEFAULTS)
        configure_axes(fig, x_axis_kwargs, y_axis_kwargs)
        figs[name] = fig
    return figs


def nested_bar():
    raise NotImplementedError


def joint_distribution():
    raise NotImplementedError


def marginal_distribution():
    raise NotImplementedError


def taylor_diagram():
    raise NotImplementedError


def probabilistic_timeseries():
    raise NotImplementedError


def reliability_diagram():
    raise NotImplementedError


def rank_histogram():
    raise NotImplementedError


def output_svg(fig):
    """
    Generates an SVG from the Plotly figure. Errors in the process are logged
    and an SVG with error text is returned.

    Parameters
    ----------
    fig : plotly.graph_objects.Figure

    Returns
    -------
    svg : str

    Notes
    -----
    Requires `Orca <https://plot.ly/python/orca-management/>`_ for generating
    svgs. If orca is not installed, an svg with an error message will be
    returned.
    """
    try:
        svg = fig.to_image(format='svg').decode('utf-8')
    except Exception:
        try:
            name = fig.layout.title['text'][3:-4]
        except Exception:
            name = 'unnamed'
        logger.error('Could not generate SVG for figure %s', name)
        svg = (
            '<svg width="100%" height="100%">'
            '<text x="50" y="50" class="alert alert-error">'
            'Unable to generate SVG plot.'
            '</text>'
            '</svg>')
    return svg


def output_pdf(fig):
    """
    Generates an PDF from the Plotly figure. Errors in the process are logged
    and an PDF with error text is returned.

    Parameters
    ----------
    fig : plotly.graph_objects.Figure

    Returns
    -------
    pdf : str
       An ASCII-85 encoded PDF

    Notes
    -----
    Requires `Orca <https://plot.ly/python/orca-management/>`_ for generating
    pdfs. If orca is not installed, an pdf with an error message will be
    returned.
    """
    try:
        pdf = base64.a85encode(fig.to_image(format='pdf')).decode('utf-8')
    except Exception:
        try:
            name = fig.layout.title['text'][3:-4]
        except Exception:
            name = 'unnamed'
        logger.error('Could not generate PDF for figure %s', name)
        # should have same text as fail SVG
        pdf = fail_pdf
    return pdf


def raw_report_plots(report, metrics):
    """Create a RawReportPlots object from the metrics of a report.

    Parameters
    ----------
    report: :py:class:`solarforecastarbiter.datamodel.Report`
    metrics: tuple of :py:class:`solarforecastarbiter.datamodel.MetricResult`

    Returns
    -------
    :py:class:`solarforecastarbiter.datamodel.RawReportPlots`
    """
    metrics_df = construct_metrics_dataframe(metrics, rename=abbreviate)
    # Create initial bar figures
    figure_dict = {}
    # Components for other metrics
    for category in report.report_parameters.categories:
        for metric in report.report_parameters.metrics:
            if category == 'total':
                fig = bar(metrics_df, metric)
                figure_dict[f'total::{metric}::all'] = fig
            else:
                figs = bar_subdivisions(metrics_df, category, metric)
                for name, fig in figs.items():
                    figure_dict[f'{category}::{metric}::{name}'] = fig
    mplots = []

    for k, v in figure_dict.items():
        cat, met, name = k.split('::', 2)
        figure_spec = v.to_json()
        pdf = output_pdf(v)
        mplots.append(datamodel.PlotlyReportFigure(
            name=name, category=cat, metric=met, spec=figure_spec,
            pdf=pdf, figure_type='bar'))

    out = datamodel.RawReportPlots(tuple(mplots), plotly_version)
    return out


def timeseries_plots(report):
    """Return the components for timeseries and scatter plots of the
    processed forecasts and observations.

    Parameters
    ----------
    report: :py:class:`solarforecastarbiter.datamodel.Report`

    Returns
    -------
    timeseries_spec: str
        String json specification of the timeseries plot
    scatter_spec: the
        String json specification of the scatter plot
    timeseries_prob_spec: None or str
        If report contains a probabilistic forecast with axis='x',
        string json specification of the probability vs. time plot.
        Otherwise None.
    includes_distribution: bool
        True if the a plot was created for a pair containing a
        ProbabilisticForecast.
    """
    value_df, meta_df = construct_timeseries_dataframe(report)
    pfxobs = report.raw_report.processed_forecasts_observations
    units = pfxobs[0].original.forecast.units
    units = units.replace('^2', '<sup>2</sup>')

    # data (units) vs time plot for the observation, deterministic fx,
    # and y-axis probabilistic fx
    ts_fig = timeseries(
        value_df, meta_df, report.report_parameters.start,
        report.report_parameters.end, units, (None, 'y'),
        report.raw_report.timezone)
    ts_fig.update_layout(
        plot_bgcolor=PLOT_BGCOLOR,
        font=dict(size=14),
        margin=PLOT_MARGINS,
    )

    # probability vs time plot for the x-axis probabilistic fx
    if any(
            (
                isinstance(pfxob.original.forecast, (
                    datamodel.ProbabilisticForecast,
                    datamodel.ProbabilisticForecastConstantValue)) and
                pfxob.original.forecast.axis == 'x')
            for pfxob in pfxobs
            ):
        ts_prob_fig = timeseries(
            value_df, meta_df, report.report_parameters.start,
            report.report_parameters.end, units, ('x',),
            report.raw_report.timezone)
        ts_prob_fig.update_layout(
            plot_bgcolor=PLOT_BGCOLOR,
            font=dict(size=14),
            margin=PLOT_MARGINS,
        )
        ts_prob_fig_json = ts_prob_fig.to_json()
    else:
        ts_prob_fig_json = None

    # switch secondary plot based on forecast type
    pfxobs = report.raw_report.processed_forecasts_observations
    fx = pfxobs[0].original.forecast
    if isinstance(fx, datamodel.EventForecast):
        scat_fig = event_histogram(value_df, meta_df)
        scat_fig.update_layout(
            plot_bgcolor=PLOT_BGCOLOR,
            font=dict(size=14),
            margin=PLOT_MARGINS,
        )
    else:
        margin = PLOT_MARGINS.copy()
        margin.pop('pad', None)
        scat_fig = scatter(value_df, meta_df, units)
        scat_fig.update_layout(
            plot_bgcolor=PLOT_BGCOLOR,
            font=dict(size=14),
            width=700,
            height=500,
            autosize=False,
            xaxis=dict(scaleanchor="y", scaleratio=1, constrain="domain"),
            yaxis=dict(constrain="domain"),
            margin=margin,
        )
    includes_distribution = any(
        (
            isinstance(pfxob.original.forecast,
                       datamodel.ProbabilisticForecast) and
            pfxob.original.forecast.axis == 'y')
        for pfxob in pfxobs)
    return (ts_fig.to_json(), scat_fig.to_json(), ts_prob_fig_json,
            includes_distribution)



import copy
import inspect
import json
import os


import numpy as np
import pandas as pd
import pytest


from solarforecastarbiter.datamodel import Site, Observation


TEST_DATA_DIR = os.path.dirname(
    os.path.abspath(inspect.getfile(inspect.currentframe())))


def site_dicts():
    return [copy.deepcopy(site) for site in [
        {
            'name': 'site',
            'latitude': 1,
            'longitude': 1,
            'elevation': 5,
            'timezone': 'Etc/GMT+8',
            'extra_parameters': {
                "network": "DOE ARM",
                "network_api_id": 'B13',
                "network_api_abbreviation": 'sgp',
                "observation_interval_length": 1,
                "datastreams": {
                    'qcrad': 'sgpqcradlong1E13.c1',
                    'met': {
                        'sgpmetE13.b1': '2019-01-01/2020-01-01',  # noqa
                        'sgpmetE13.b2': '2020-01-01/2021-01-01',  # noqa
                    },
                },
            },
        },
        {
            'name': 'site2',
            'latitude': 2,
            'longitude': 2,
            'elevation': 5,
            'timezone': 'Etc/GMT+8',
            'extra_parameters': {"network": "NOAA SURFRAD",
                                 "network_api_id": 'some_id',
                                 "network_api_abbreviation": 'abbrv',
                                 "observation_interval_length": 5},
        },
        {
            'name': 'site3',
            'latitude': 3,
            'longitude': -3,
            'elevation': 6,
            'timezone': 'Etc/GMT+8',
            'extra_parameters': {"network": "NOAA SOLRAD",
                                 "network_api_id": 'some_id',
                                 "network_api_abbreviation": 'abbrv',
                                 "observation_interval_length": 1},
        },
        {
            'name': 'site4',
            'latitude': 4,
            'longitude': -5,
            'elevation': 12,
            'timezone': 'Etc/GMT+8',
            'extra_parameters': {"observation_interval_length": 1,
                                 "network": 'NREL MIDC',
                                 "network_api_id": 'BMS',
                                 "network_api_abbreviation": 'abbrv'},
        },
        {
            'name': 'site4',
            'latitude': 4,
            'longitude': -5,
            'elevation': 12,
            'timezone': 'Etc/GMT+8',
            'extra_parameters': {"observation_interval_length": 1,
                                 "network": 'Unincorporated',
                                 "network_api_id": 'BMS',
                                 "network_api_abbreviation": 'abbrv'},
        }
    ]]


def expected_site(site):
    new_site = site.copy()
    network = site['extra_parameters'].get('network', '')
    new_site['name'] = f"{network} {site['name']}"
    new_site.update({'extra_parameters': json.dumps(site['extra_parameters'])})
    return new_site


site_string_dicts = [expected_site(site) for site in site_dicts()]
site_objects = [Site.from_dict(site) for site in site_string_dicts]


@pytest.fixture
def site_dicts_param():
    return site_string_dicts


@pytest.fixture
def site_objects_param():
    return site_objects


def site_to_obs(site):
    ep = json.loads(site.extra_parameters)
    interval_length = ep['observation_interval_length']
    return Observation.from_dict({
        'name': 'site ghi',
        'variable': 'ghi',
        'interval_label': 'ending',
        'interval_value_type': 'interval_mean',
        'interval_length': interval_length,
        'site': site,
        'uncertainty': 0,
        'extra_parameters': site.extra_parameters
    })


@pytest.fixture
def observation_objects_param(site_objects_param):
    return [site_to_obs(site) for site in site_objects_param]


@pytest.fixture
def networks():
    return ['DOE ARM', 'NOAA SURFRAD', 'NOAA SOLRAD', 'Unincorporated']


@pytest.fixture
def mock_api(mocker, site_objects_param, observation_objects_param):
    api = mocker.MagicMock()
    api.list_sites.return_value = site_objects_param
    api.list_observations.return_value = observation_objects_param
    return api


index = pd.date_range('20190101T1200Z', '20190101T1229Z',
                      freq='min', tz='UTC')
values = np.arange(100, 130)


@pytest.fixture
def start():
    return pd.Timestamp('20190101T1200Z')


@pytest.fixture
def end():
    return pd.Timestamp('20190101T1229Z')


@pytest.fixture
def fake_ghi_data():
    df = pd.DataFrame(index=index, data={'ghi': values})
    df['quality_flag'] = 0
    return df


@pytest.fixture
def mock_fetch(mocker, fake_ghi_data):
    fetch = mocker.MagicMock()
    fetch.return_value = fake_ghi_data
    return fetch


@pytest.fixture
def test_json_site_file():
    return os.path.join(TEST_DATA_DIR, 'data', 'test_site.json')



from functools import partial
import json
import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest

from solarforecastarbiter.io import utils

# data for test Dataframe
TEST_DICT = {'value': [2.0, 43.9, 338.0, -199.7, 0.32],
             'quality_flag': [1, 1, 9, 5, 2]}

DF_INDEX = pd.date_range(start=pd.Timestamp('2019-01-24T00:00'),
                         freq='1min',
                         periods=5,
                         tz='UTC', name='timestamp')
TEST_DATA = pd.DataFrame(TEST_DICT, index=DF_INDEX)
EMPTY_SERIES = pd.Series(dtype=float)
EMPTY_TIMESERIES = pd.Series([], name='value', index=pd.DatetimeIndex(
    [], name='timestamp', tz='UTC'), dtype=float)

EMPTY_DATAFRAME = pd.DataFrame(dtype=float)
EMPTY_TIME_DATAFRAME = pd.DataFrame([], index=pd.DatetimeIndex(
    [], name='timestamp', tz='UTC'), dtype=float)
TEST_DATAFRAME = pd.DataFrame({
    '25.0': [0.0, 1, 2, 3, 4, 5],
    '50.0': [1.0, 2, 3, 4, 5, 6],
    '75.0': [2.0, 3, 4, 5, 6, 7]},
    index=pd.date_range(start='20190101T0600',
                        end='20190101T1100',
                        freq='1h',
                        tz='America/Denver',
                        name='timestamp')).tz_convert('UTC')


@pytest.mark.parametrize('dump_quality,default_flag,flag_value', [
    (False, None, 1),
    (True, 2, 2)
])
def test_obs_df_to_json(dump_quality, default_flag, flag_value):
    td = TEST_DATA.copy()
    if dump_quality:
        del td['quality_flag']
    converted = utils.observation_df_to_json_payload(td, default_flag)
    converted_dict = json.loads(converted)
    assert 'values' in converted_dict
    values = converted_dict['values']
    assert len(values) == 5
    assert values[0]['timestamp'] == '2019-01-24T00:00:00Z'
    assert values[0]['quality_flag'] == flag_value
    assert isinstance(values[0]['value'], float)


def test_obs_df_to_json_no_quality():
    td = TEST_DATA.copy()
    del td['quality_flag']
    with pytest.raises(KeyError):
        utils.observation_df_to_json_payload(td)


def test_obs_df_to_json_no_values():
    td = TEST_DATA.copy().rename(columns={'value': 'val1'})
    with pytest.raises(KeyError):
        utils.observation_df_to_json_payload(td)


def test_forecast_series_to_json():
    series = pd.Series([0, 1, 2, 3, 4], index=pd.date_range(
        start='2019-01-01T12:00Z', freq='5min', periods=5))
    expected = [{'value': 0.0, 'timestamp': '2019-01-01T12:00:00Z'},
                {'value': 1.0, 'timestamp': '2019-01-01T12:05:00Z'},
                {'value': 2.0, 'timestamp': '2019-01-01T12:10:00Z'},
                {'value': 3.0, 'timestamp': '2019-01-01T12:15:00Z'},
                {'value': 4.0, 'timestamp': '2019-01-01T12:20:00Z'}]
    json_out = utils.forecast_object_to_json(series)
    assert json.loads(json_out)['values'] == expected


def test_json_payload_to_observation_df(observation_values,
                                        observation_values_text):
    out = utils.json_payload_to_observation_df(
        json.loads(observation_values_text))
    pdt.assert_frame_equal(out, observation_values)


def test_json_payload_to_forecast_series(forecast_values,
                                         forecast_values_text):
    out = utils.json_payload_to_forecast_series(
        json.loads(forecast_values_text))
    pdt.assert_series_equal(out, forecast_values)


def test_empty_payload_to_obsevation_df():
    out = utils.json_payload_to_observation_df({'values': []})
    assert set(out.columns) == {'value', 'quality_flag'}
    assert isinstance(out.index, pd.DatetimeIndex)


def test_empty_payload_to_forecast_series():
    out = utils.json_payload_to_forecast_series({'values': []})
    assert isinstance(out.index, pd.DatetimeIndex)


def test_null_json_payload_to_observation_df():
    observation_values_text = b"""
{
  "_links": {
    "metadata": ""
  },
  "observation_id": "OBSID",
  "values": [
    {
      "quality_flag": 1,
      "timestamp": "2019-01-01T12:00:00-0700",
      "value": null
    },
    {
      "quality_flag": 1,
      "timestamp": "2019-01-01T12:05:00-0700",
      "value": null
    }
  ]
}"""
    ind = pd.DatetimeIndex([
        pd.Timestamp("2019-01-01T19:00:00Z"),
        pd.Timestamp("2019-01-01T19:05:00Z")
    ], name='timestamp')
    observation_values = pd.DataFrame({
        'value': pd.Series([None, None], index=ind, dtype=float),
        'quality_flag': pd.Series([1, 1], index=ind)
    })
    out = utils.json_payload_to_observation_df(
        json.loads(observation_values_text))
    pdt.assert_frame_equal(out, observation_values)


def test_null_json_payload_to_forecast_series():
    forecast_values_text = b"""
{
  "_links": {
    "metadata": ""
  },
  "forecast_id": "OBSID",
  "values": [
    {
      "timestamp": "2019-01-01T12:00:00-0700",
      "value": null
    },
    {
      "timestamp": "2019-01-01T12:05:00-0700",
      "value": null
    }
  ]
}"""
    ind = pd.DatetimeIndex([
        pd.Timestamp("2019-01-01T19:00:00Z"),
        pd.Timestamp("2019-01-01T19:05:00Z")
    ], name='timestamp')
    forecast_values = pd.Series([None, None], index=ind, dtype=float,
                                name='value')
    out = utils.json_payload_to_forecast_series(
        json.loads(forecast_values_text))
    pdt.assert_series_equal(out, forecast_values)


@pytest.mark.parametrize('label,exp,start,end', [
    ('instant', TEST_DATA, None, None),
    (None, TEST_DATA, None, None),
    ('ending', TEST_DATA.iloc[1:], None, None),
    ('beginning', TEST_DATA.iloc[:-1], None, None),
    pytest.param('er', TEST_DATA, None, None,
                 marks=pytest.mark.xfail(raises=ValueError)),
    # start/end outside data
    ('ending', TEST_DATA, pd.Timestamp('20190123T2300Z'), None),
    ('beginning', TEST_DATA, None, pd.Timestamp('20190124T0100Z')),
    # more limited
    ('ending', TEST_DATA.iloc[2:], pd.Timestamp('20190124T0001Z'), None),
    ('beginning', TEST_DATA.iloc[:-2], None,
     pd.Timestamp('20190124T0003Z')),
    ('instant', TEST_DATA.iloc[1:-1], pd.Timestamp('20190124T0001Z'),
     pd.Timestamp('20190124T0003Z')),
])
def test_adjust_timeseries_for_interval_label(label, exp, start, end):
    start = start or pd.Timestamp('2019-01-24T00:00Z')
    end = end or pd.Timestamp('2019-01-24T00:04Z')
    out = utils.adjust_timeseries_for_interval_label(
        TEST_DATA, label, start, end)
    pdt.assert_frame_equal(exp, out)


def test_adjust_timeseries_for_interval_label_no_tz():
    test_data = TEST_DATA.tz_localize(None)
    label = None
    start = pd.Timestamp('2019-01-24T00:00Z')
    end = pd.Timestamp('2019-01-24T00:04Z')
    with pytest.raises(ValueError):
        utils.adjust_timeseries_for_interval_label(
            test_data, label, start, end)


def test_adjust_timeseries_for_interval_label_no_tz_empty():
    test_data = pd.DataFrame()
    label = None
    start = pd.Timestamp('2019-01-24T00:00Z')
    end = pd.Timestamp('2019-01-24T00:04Z')
    out = utils.adjust_timeseries_for_interval_label(
            test_data, label, start, end)
    pdt.assert_frame_equal(test_data, out)


@pytest.mark.parametrize('label,exp', [
    ('instant', TEST_DATA['value']),
    ('ending', TEST_DATA['value'].iloc[1:]),
    ('beginning', TEST_DATA['value'].iloc[:-1])
])
def test_adjust_timeseries_for_interval_label_series(label, exp):
    start = pd.Timestamp('2019-01-24T00:00Z')
    end = pd.Timestamp('2019-01-24T00:04Z')
    out = utils.adjust_timeseries_for_interval_label(
        TEST_DATA['value'], label, start, end)
    pdt.assert_series_equal(exp, out)


@pytest.mark.parametrize('inp,exp', [
    (TEST_DATA['value'], '{"schema":{"version": 1, "orient": "records", "timezone": "UTC", "column": "value", "index": "timestamp", "dtype": "float64", "objtype": "Series"},"data":[{"timestamp":"2019-01-24T00:00:00Z","value":2.0},{"timestamp":"2019-01-24T00:01:00Z","value":43.9},{"timestamp":"2019-01-24T00:02:00Z","value":338.0},{"timestamp":"2019-01-24T00:03:00Z","value":-199.7},{"timestamp":"2019-01-24T00:04:00Z","value":0.32}]}'),  # NOQA: E501
    (EMPTY_TIMESERIES, '{"schema":{"version": 1, "orient": "records", "timezone": "UTC", "column": "value", "index": "timestamp", "dtype": "float64", "objtype": "Series"},"data":[]}'),  # NOQA: E501
    pytest.param(EMPTY_SERIES, '', marks=[
        pytest.mark.xfail(strict=True, type=TypeError)]),
    pytest.param(pd.Series([], dtype=float, index=pd.DatetimeIndex([])), '',
                 marks=[pytest.mark.xfail(strict=True, type=TypeError)]),
    (TEST_DATAFRAME, '{"schema":{"version": 1, "orient": "records", "timezone": "UTC", "column": ["25.0", "50.0", "75.0"], "index": "timestamp", "dtype": ["float64", "float64", "float64"], "objtype": "DataFrame"},"data":[{"timestamp":"2019-01-01T13:00:00Z","25.0":0.0,"50.0":1.0,"75.0":2.0},{"timestamp":"2019-01-01T14:00:00Z","25.0":1.0,"50.0":2.0,"75.0":3.0},{"timestamp":"2019-01-01T15:00:00Z","25.0":2.0,"50.0":3.0,"75.0":4.0},{"timestamp":"2019-01-01T16:00:00Z","25.0":3.0,"50.0":4.0,"75.0":5.0},{"timestamp":"2019-01-01T17:00:00Z","25.0":4.0,"50.0":5.0,"75.0":6.0},{"timestamp":"2019-01-01T18:00:00Z","25.0":5.0,"50.0":6.0,"75.0":7.0}]}'),  # NOQA: E501
    (EMPTY_TIME_DATAFRAME, '{"schema":{"version": 1, "orient": "records", "timezone": "UTC", "column": [], "index": "timestamp", "dtype": [], "objtype": "DataFrame"},"data":[]}'),  # NOQA: E501
    pytest.param(EMPTY_DATAFRAME, '', marks=[
        pytest.mark.xfail(strict=True, type=TypeError)]),
])
def test_serialize_timeseries(inp, exp):
    out = utils.serialize_timeseries(inp)
    outd = json.loads(out)
    assert 'schema' in outd
    assert 'data' in outd
    assert out == exp


@pytest.mark.parametrize('inp,exp', [
    ('{"schema": {"version": 0, "orient": "records", "timezone": "UTC", "column": "value", "index": "timestamp", "dtype": "float64"}, "data": []}',  # NOQA
     EMPTY_TIMESERIES),
    ('{"schema": {"version": 0, "orient": "records", "timezone": "US/Arizona", "column": "value", "index": "timestamp", "dtype": "float64"}, "data": []}',  # NOQA
     pd.Series([], name='value', index=pd.DatetimeIndex(
         [], tz='US/Arizona', name='timestamp'), dtype=float)),
    ('{"schema": {"version": 0, "orient": "records", "timezone": "UTC", "column": "value", "index": "timestamp", "dtype": "float64"}, "data": [], "other_stuff": {}}',  # NOQA
     EMPTY_TIMESERIES),
    ('{"schema": {"version": 0, "orient": "records", "timezone": "UTC", "column": "alue", "index": "timestamp", "dtype": "float64"}, "more": [], "data": []}',  # NOQA
     pd.Series([], name='alue', index=pd.DatetimeIndex(
         [], tz='UTC', name='timestamp'), dtype=float)),
    ('{"schema": {"version": 0, "orient": "records", "timezone": "UTC", "column": "value", "index": "timestamp", "dtype": "float64"}, "more": [], "data": [], "other": []}',  # NOQA
     EMPTY_TIMESERIES),
    ('{"schema": {"version": 0, "orient": "records", "timezone": "UTC", "column": "value", "index": "timestamp", "dtype": "float64"}, "data": [{"timestamp": "2019-01-01T00:00:00Z", "value": 1.0}], "other_stuff": {}}',  # NOQA
     pd.Series([1.0], index=pd.DatetimeIndex(["2019-01-01T00:00:00"],
                                             tz='UTC', name='timestamp'),
               name='value')),
    ('{"schema": {"version": 0, "orient": "records", "timezone": "UTC", "column": "value", "index": "timestamp", "dtype": "float64"}, "data": [{"timestamp": "2019-01-01T00:00:00", "value": 1.0}], "other_stuff": {}}',  # NOQA
     pd.Series([1.0], index=pd.DatetimeIndex(["2019-01-01T00:00:00"],
                                             tz='UTC', name='timestamp'),
               name='value')),
    ('{"schema": {"version": 0, "orient": "records", "timezone": "Etc/GMT+8", "column": "value", "index": "timestamp", "dtype": "float64"}, "data": [{"timestamp": "2019-01-01T00:00:00", "value": 1.0}], "other_stuff": {}}',  # NOQA
     pd.Series([1.0], index=pd.DatetimeIndex(["2019-01-01T00:00:00"],
                                             tz='Etc/GMT+8', name='timestamp'),
               name='value')),
    pytest.param(
        '{"data": [], "other_stuff": {}}',
        EMPTY_SERIES,
        marks=[pytest.mark.xfail(strict=True, type=ValueError)]),
    pytest.param(
        '{"schema": {"version": 0, "orient": "records", "timezone": "UTC", "column": "value", "index": "timestamp", "dtype": "float64"}, "other_stuff": {}}',  # NOQA
        EMPTY_SERIES,
        marks=[pytest.mark.xfail(strict=True, type=ValueError)]),
    pytest.param(
        '{"schema": {"version": 0, "timezone": "UTC", "column": "value", "index": "timestamp", "dtype": "float64"}, "data": []}',  # NOQA
        EMPTY_SERIES,
        marks=[pytest.mark.xfail(strict=True, type=KeyError)]),
    ('{"schema": {"version": 1, "objtype": "Series", "orient": "records", "timezone": "UTC", "column": "value", "index": "timestamp", "dtype": "float64"}, "data": []}',  # NOQA
     EMPTY_TIMESERIES)
])
def test_deserialize_timeseries(inp, exp):
    out = utils.deserialize_timeseries(inp)
    pdt.assert_series_equal(out, exp)


@pytest.mark.parametrize('inp,exp', [
    ('{"schema":{"version": 1, "orient": "records", "timezone": "UTC", "column": [], "index": "timestamp", "dtype": [], "objtype": "DataFrame"},"data":[]}',  # NOQA
     EMPTY_TIME_DATAFRAME),
    ('{"schema": {"version": 1, "objtype": "DataFrame", "orient": "records", "timezone": "UTC", "column": [], "index": "timestamp", "dtype": ["float64"]}, "data": []}',  # NOQA
     EMPTY_TIME_DATAFRAME),
    ('{"schema": {"version": 1, "objtype": "DataFrame", "orient": "records", "timezone": "UTC", "column": ["25.0"], "index": "timestamp", "dtype": ["float64"]}, "data": [{"timestamp": "2019-01-01T00:00:00", "25.0": 1.0}], "other_stuff": {}}',  # NOQA
     pd.DataFrame({'25.0': 1.0}, index=pd.DatetimeIndex(
         ["2019-01-01T00:00:00"], tz='UTC', name='timestamp'))),
    ('{"schema": {"version": 1, "objtype": "DataFrame", "orient": "records", "timezone": "Etc/GMT+8", "column": ["25.0"], "index": "timestamp", "dtype": ["float64"]}, "data": [{"timestamp": "2019-01-01T00:00:00", "25.0": 1.0}], "other_stuff": {}}',  # NOQA
     pd.DataFrame({'25.0': 1.0}, index=pd.DatetimeIndex(
         ["2019-01-01T00:00:00"], tz='Etc/GMT+8', name='timestamp'))),
])
def test_deserialize_timeseries_frame(inp, exp):
    out = utils.deserialize_timeseries(inp)
    pdt.assert_frame_equal(out, exp)


def test_serialize_roundtrip():
    ser = utils.serialize_timeseries(TEST_DATA['value'])
    out = utils.deserialize_timeseries(ser)
    pdt.assert_series_equal(out, TEST_DATA['value'])


# use the conftest.py dataframe for security against refactoring
def test_serialize_roundtrip_frame(prob_forecast_values):
    ser = utils.serialize_timeseries(prob_forecast_values)
    out = utils.deserialize_timeseries(ser)
    pdt.assert_frame_equal(out, prob_forecast_values)


def test_serialize_roundtrip_frame_floats(prob_forecast_values):
    # check that roundtrip is not faithful if input columns is Float64Index
    prob_forecast_floats = prob_forecast_values.copy()
    prob_forecast_floats.columns = prob_forecast_floats.columns.astype(float)
    ser = utils.serialize_timeseries(prob_forecast_floats)
    out = utils.deserialize_timeseries(ser)
    pdt.assert_frame_equal(out, prob_forecast_values)


def test_hidden_token():
    ht = utils.HiddenToken('THETOKEN')
    assert str(ht) != 'THETOKEN'
    assert ht.token == 'THETOKEN'


@pytest.fixture(params=[0, 1, 2, 3])
def start_end(request):
    if request.param == 0:
        return (pd.Timestamp('2019-01-01T12:00:00Z'),
                pd.Timestamp('2019-01-12T17:47:22-0700'))
    elif request.param == 1:
        return ('2019-01-01T12:00:00Z',
                pd.Timestamp('2019-01-12T17:47:22-0700'))
    elif request.param == 2:
        return ('2019-01-01T12:00:00Z',
                '2019-01-12T17:47:22-0700')
    elif request.param == 3:
        return (pd.Timestamp('2019-01-01T12:00:00Z'),
                '2019-01-12T17:47:22-0700')


@pytest.fixture(params=[0, 1, 2, 3, 4])
def tfunc(request, start_end):
    start, end = start_end
    if request.param == 0:
        def f(start, end): return start, end
        return partial(f, start, end)
    elif request.param == 1:
        def f(start, end): return start, end
        return partial(f, start=start, end=end)
    elif request.param == 2:
        def f(start, *, end): return start, end
        return partial(f, start, end=end)
    elif request.param == 3:
        def f(te, start, more=None, *, end): return start, end
        return partial(f, '', start, end=end)
    elif request.param == 4:
        def f(te, startitat, end, more=None, start=None): return start, end
        return partial(f, '', 'now', start=start, end=end)


def test_ensure_timestamps_partials(tfunc):
    dec_f = utils.ensure_timestamps('start', 'end')(tfunc.func)
    s, e = dec_f(*tfunc.args, **tfunc.keywords)
    assert s == pd.Timestamp('2019-01-01T12:00:00Z')
    assert e == pd.Timestamp('2019-01-12T17:47:22-0700')


def test_ensure_timestamps_normal(start_end):
    @utils.ensure_timestamps('start', 'end')
    def f(other, start, end, x=None):
        """cool docstring"""
        return start, end
    s, e = start_end
    start, end = f('', s, e)
    assert f.__doc__ == 'cool docstring'
    assert start == pd.Timestamp('2019-01-01T12:00:00Z')
    assert end == pd.Timestamp('2019-01-12T17:47:22-0700')


def test_ensure_timestamps_other_args():
    @utils.ensure_timestamps('x')
    def f(other, start, end, x=None):
        """cool docstring"""
        return start, end, x
    start, end, x = f('', 'a', 'end', '2019-01-01T12:00:00Z')
    assert f.__doc__ == 'cool docstring'
    assert x == pd.Timestamp('2019-01-01T12:00:00Z')
    assert start == 'a'
    assert end == 'end'


def test_ensure_timestamps_err():
    @utils.ensure_timestamps('start', 'end', 'x')
    def f(other, start, end, x=None):  # pragma: no cover
        return start, end
    with pytest.raises(ValueError):
        start, end = f('', '2019-09-01T12:00Z', 'blah')


def test_load_report_values(raw_report, report_objects):
    _, obs, fx0, fx1, agg, fxagg = report_objects
    ser = pd.Series(np.random.random(10),
                    name='value', index=pd.date_range(
                        start='20200101', freq='5min', periods=10,
                        tz='UTC', name='timestamp'))
    val = utils.serialize_timeseries(ser)
    vals = [{'id': id_, 'processed_values': val} for id_ in
            (fx0.forecast_id, fx1.forecast_id, obs.observation_id,
             agg.aggregate_id, fxagg.forecast_id)]
    inp = raw_report(False)
    out = utils.load_report_values(inp, vals)
    for fxo in out:
        pdt.assert_series_equal(fxo.forecast_values, ser)
        pdt.assert_series_equal(fxo.observation_values, ser)



# maybe rename nwp.py or models_nwp.py
"""
Default processing functions for data from NOAA weather models.

All public functions in this module have the same signature.

The functions accept:

  * latitude : float
  * longitude : float
  * elevation : float
  * init_time : pd.Timestamp
      Full datetime of a model initialization
  * start : pd.Timestamp
  * end : pd.Timestamp
  * load_forecast : function
      A function that accepts the arguments above and returns the
      correct data. Enables users to supply their own data independently
      of the `solarforecastarbiter.io` module.
  * __model : str
      The NWP model that the processing function is associated with.

The functions return a tuple of:

  * ghi : pd.Series
  * dni : pd.Series
  * dhi : pd.Series
  * air_temperature : pd.Series
  * wind_speed : pd.Series
  * resampler : function
      A function that resamples data to the appropriate frequency.
      This function is to be applied to all forecast variables after
      power is calculated.
  * solar_position_calculator : function
      A function that returns solar position at the ghi forecast times
      (after internal interpolation, before external resampling). The
      function will return results immediately if solar position is
      already known or will call a solar position calculation algorithm
      and then return.

Most of the model functions return forecast data interpolated to 5
minute frequency. Interpolation to 5 minutes reduces the errors
associated with solar position and irradiance to power models (these
models assume instantaneous inputs). Why 5 minutes? It's a round number
that produces about 10 data points per hour, so it's reasonable for hour
average calculations. It is expected that after calculating power, users
will apply the `resampler` function to both the weather and power
forecasts. These functions may be most useful if the user would like to
understand the performance of a NWP model with modest post-processing.

Several model functions return instantaneous data that is coincident
with the NWP model forecast time (15 minutes or hourly, depending on the
NWP model). The resamplers returned by these functions do not modify the
data (though they do define the frequency attribute of the data's
DatetimeIndex). These functions may most useful if the user would like
to understand the raw performance of a NWP model.

The functions in this module accept primitives (floats, strings, etc.)
rather than objects defined in :py:mod:`solarforecastarbiter.datamodel`
because we anticipate that these functions may be of more general use
and that functions that accept primitives may be easier to maintain in
the long run.
"""
from functools import partial
import inspect


from solarforecastarbiter import datamodel, pvmodel
from solarforecastarbiter.io.nwp import load_forecast
from solarforecastarbiter.io.utils import adjust_start_end_for_interval_label
from solarforecastarbiter.reference_forecasts import forecast

import pandas as pd


def get_nwp_model(func):
    """Get the NWP model string from a modeling function"""
    return inspect.signature(func).parameters['__model'].default


def _resample_using_cloud_cover(latitude, longitude, elevation,
                                cloud_cover, air_temperature, wind_speed,
                                start, end, interval_label,
                                fill_method, solar_position=None):
    """
    Calculate all irradiance components from cloud cover.

    Cloud cover from GFS is an interval average with ending label.
    Air temperature and wind speed are instantaneous values.
    Intervals are 1, 3, or 6 hours in length.

    Cloud cover, air temperature, and wind speed from RAP and NAM are
    instantaneous.

    To accurately convert from cloud cover to irradiance, we need to
    interpolate this data to subhourly resolution because solar position
    and PV power calculations assume instantaneous inputs at each time.

    Parameters
    ----------
    latitude : float
    longitude : float
    elevation : float
    cloud_cover : pd.Series
    air_temperature : pd.Series
    wind_speed : pd.Series
    start : pd.Timestamp
    end : pd.Timestamp
    interval_label : str
        beginning, ending, or instant
    fill_method : str
        'bfill' (recommended for GFS), 'interpolate' (recommended for
        NAM/RAP), or any other method of pd.Series.
    solar_position : pd.DataFrame or None
        Provide a DataFrame to avoid unnecessary recomputation for e.g.
        GEFS members. If None, solar position is computed.

    Returns
    -------
    ghi : pd.Series
    dni : pd.Series
    dhi : pd.Series
    air_temperature : pd.Series
    wind_speed : pd.Series
    resampler : function
    sol_pos_calculator : function
        When called, immediatedly returns pre-computed solar position.
    """
    # Resample cloud cover, temp, and wind to higher temporal resolution
    # because solar position and PV power calculations assume instantaneous
    # inputs. Why 5 minutes? It's a round number that produces order 10 data
    # points per hour, so it's reasonable for hour average calculations.
    # Cloud cover should be filled backwards for GFS because model output
    # represents average over the previous hour and interpolated for RAP and
    # NAM because model outputs represent instantaneous values. Air temperature
    # and wind are interpolated because model output represents
    # instantaneous values.
    freq = '5min'
    start_adj, end_adj = adjust_start_end_for_interval_label(interval_label,
                                                             start, end)
    cloud_cover = forecast.reindex_fill_slice(
        cloud_cover, freq=freq, start=start, end=end,
        start_slice=start_adj, end_slice=end_adj,
        fill_method=fill_method)
    resample_fill_slicer = partial(
        forecast.reindex_fill_slice, freq=freq, start=start, end=end,
        start_slice=start_adj, end_slice=end_adj, fill_method='interpolate')
    air_temperature, wind_speed = [
        resample_fill_slicer(v) for v in (air_temperature, wind_speed)
    ]
    if solar_position is None:
        solar_position = pvmodel.calculate_solar_position(
            latitude, longitude, elevation, cloud_cover.index)
    ghi, dni, dhi = forecast.cloud_cover_to_irradiance(
        latitude, longitude, elevation, cloud_cover,
        solar_position['apparent_zenith'], solar_position['zenith'])

    label = datamodel.CLOSED_MAPPING[interval_label]
    resampler = partial(forecast.resample, freq='1h', label=label)

    def solar_pos_calculator(): return solar_position

    return (ghi, dni, dhi, air_temperature, wind_speed,
            resampler, solar_pos_calculator)


def _ghi_to_dni_dhi(latitude, longitude, elevation, ghi):
    """
    Calculate DNI, DHI from GHI and calculated solar position.
    """
    solar_position = pvmodel.calculate_solar_position(
        latitude, longitude, elevation, ghi.index)
    dni, dhi = pvmodel.complete_irradiance_components(
        ghi, solar_position['zenith'])

    def solar_pos_calculator(): return solar_position
    return dni, dhi, solar_pos_calculator


def hrrr_subhourly_to_subhourly_instantaneous(latitude, longitude, elevation,
                                              init_time, start, end,
                                              interval_label,
                                              load_forecast=load_forecast,
                                              *, __model='hrrr_subhourly'):
    """
    Subhourly (15 min) instantantaneous HRRR forecast.
    GHI, DNI, DHI directly from model.
    Max forecast horizon 18 or 36 hours (0Z, 6Z, 12Z, 18Z).

    Parameters
    ----------
    latitude : float
    longitude : float
    elevation : float
    init_time : pd.Timestamp
        Full datetime of a model initialization
    start : pd.Timestamp
        Forecast start. Forecast is inclusive of this point.
    end : pd.Timestamp
        Forecast end. Forecast is exclusive of this point.
    interval_label : str
        Must be instant
    """
    start_adj, end_adj = adjust_start_end_for_interval_label(
        interval_label, start, end, limit_instant=True)
    ghi, dni, dhi, air_temperature, wind_speed = load_forecast(
        latitude, longitude, init_time, start_adj, end_adj, __model)
    # resampler takes 15 min instantaneous in, retuns 15 min instantaneous out
    # still want to call resample, rather than pass through lambda x: x
    # so that DatetimeIndex has well-defined freq attribute
    resampler = partial(forecast.resample, freq='15min')
    solar_pos_calculator = partial(
        pvmodel.calculate_solar_position, latitude, longitude, elevation,
        ghi.index)
    return (ghi, dni, dhi, air_temperature, wind_speed,
            resampler, solar_pos_calculator)


def hrrr_subhourly_to_hourly_mean(latitude, longitude, elevation,
                                  init_time, start, end, interval_label,
                                  load_forecast=load_forecast,
                                  *, __model='hrrr_subhourly'):
    """
    Hourly mean HRRR forecast.
    GHI, DNI, DHI directly from model, resampled.
    Max forecast horizon 18 or 36 hours (0Z, 6Z, 12Z, 18Z).

    Parameters
    ----------
    latitude : float
    longitude : float
    elevation : float
    init_time : pd.Timestamp
        Full datetime of a model initialization
    start : pd.Timestamp
        Forecast start. Forecast is inclusive of this instant if
        interval_label is *beginning* and exclusive of this instant if
        interval_label is *ending*.
    end : pd.Timestamp
        Forecast end. Forecast is exclusive of this instant if
        interval_label is *beginning* and inclusive of this instant if
        interval_label is *ending*.
    interval_label : str
        Must be *beginning* or *ending*
    """
    ghi, dni, dhi, air_temperature, wind_speed = load_forecast(
        latitude, longitude, init_time, start, end, __model)
    # Interpolate irrad, temp, wind data to 5 min to
    # minimize weather to power errors. Either start or end is outside of
    # forecast, but is needed for subhourly interpolation. After
    # interpolation, we slice the extra point out of the interpolated
    # output.
    start_adj, end_adj = adjust_start_end_for_interval_label(interval_label,
                                                             start, end)
    resample_interpolate_slicer = partial(forecast.reindex_fill_slice,
                                          freq='5min', start_slice=start_adj,
                                          end_slice=end_adj)
    ghi, dni, dhi, air_temperature, wind_speed = [
        resample_interpolate_slicer(v) for v in
        (ghi, dni, dhi, air_temperature, wind_speed)
    ]
    # weather (and optionally power) will eventually be resampled
    # to hourly average using resampler defined below
    label = datamodel.CLOSED_MAPPING[interval_label]
    resampler = partial(forecast.resample, freq='1h', label=label)
    solar_pos_calculator = partial(
        pvmodel.calculate_solar_position, latitude, longitude, elevation,
        ghi.index)
    return (ghi, dni, dhi, air_temperature, wind_speed,
            resampler, solar_pos_calculator)


def rap_ghi_to_instantaneous(latitude, longitude, elevation,
                             init_time, start, end, interval_label,
                             load_forecast=load_forecast,
                             *, __model='rap'):
    """
    Hourly instantantaneous RAP forecast.
    GHI directly from NWP model. DNI, DHI computed.
    Max forecast horizon 21 or 39 (3Z, 9Z, 15Z, 21Z) hours.

    Parameters
    ----------
    latitude : float
    longitude : float
    elevation : float
    init_time : pd.Timestamp
        Full datetime of a model initialization
    start : pd.Timestamp
        Forecast start. Forecast is inclusive of this point.
    end : pd.Timestamp
        Forecast end. Forecast is exclusive of this point.
    interval_label : str
        Must be instant
    """
    # ghi dni and dhi not in RAP output available from g2sub service
    ghi, air_temperature, wind_speed = load_forecast(
        latitude, longitude, init_time, start, end, __model,
        variables=('ghi', 'air_temperature', 'wind_speed'))
    dni, dhi, solar_pos_calculator = _ghi_to_dni_dhi(
        latitude, longitude, elevation, ghi)
    # hourly instant in, hourly instant out
    resampler = partial(forecast.resample, freq='1h')
    return (ghi, dni, dhi, air_temperature, wind_speed,
            resampler, solar_pos_calculator)


def rap_cloud_cover_to_hourly_mean(latitude, longitude, elevation,
                                   init_time, start, end, interval_label,
                                   load_forecast=load_forecast,
                                   *, __model='rap'):
    """
    Take hourly RAP instantantaneous cloud cover and convert it to
    hourly average forecasts.
    GHI from NWP model cloud cover. DNI, DHI computed.
    Max forecast horizon 21 or 39 (3Z, 9Z, 15Z, 21Z) hours.

    Parameters
    ----------
    latitude : float
    longitude : float
    elevation : float
    init_time : pd.Timestamp
        Full datetime of a model initialization
    start : pd.Timestamp
        Forecast start. Forecast is inclusive of this instant if
        interval_label is *beginning* and exclusive of this instant if
        interval_label is *ending*.
    end : pd.Timestamp
        Forecast end. Forecast is exclusive of this instant if
        interval_label is *beginning* and inclusive of this instant if
        interval_label is *ending*.
    interval_label : str
        Must be *beginning* or *ending*
    """
    cloud_cover, air_temperature, wind_speed = load_forecast(
        latitude, longitude, init_time, start, end, __model,
        variables=('cloud_cover', 'air_temperature', 'wind_speed'))
    return _resample_using_cloud_cover(latitude, longitude, elevation,
                                       cloud_cover, air_temperature,
                                       wind_speed, start, end, interval_label,
                                       'interpolate')


def gfs_quarter_deg_3hour_to_hourly_mean(latitude, longitude, elevation,
                                         init_time, start, end, interval_label,
                                         load_forecast=load_forecast,
                                         *, __model='gfs_3h'):
    """
    Take 3 hr GFS and convert it to hourly average data.
    GHI from NWP model cloud cover. DNI, DHI computed.
    Max forecast horizon 240 hours.

    Parameters
    ----------
    latitude : float
    longitude : float
    elevation : float
    init_time : pd.Timestamp
        Full datetime of a model initialization
    start : pd.Timestamp
        Forecast start. Forecast is inclusive of this instant if
        interval_label is *beginning* and exclusive of this instant if
        interval_label is *ending*.
    end : pd.Timestamp
        Forecast end. Forecast is exclusive of this instant if
        interval_label is *beginning* and inclusive of this instant if
        interval_label is *ending*.
    interval_label : str
        Must be *beginning* or *ending*
    """
    start_floored, end_ceil = _adjust_gfs_start_end(start, end)
    cloud_cover_mixed, air_temperature, wind_speed = load_forecast(
        latitude, longitude, init_time, start_floored, end_ceil, __model,
        variables=('cloud_cover', 'air_temperature', 'wind_speed'))
    cloud_cover = forecast.unmix_intervals(cloud_cover_mixed)
    return _resample_using_cloud_cover(latitude, longitude, elevation,
                                       cloud_cover, air_temperature,
                                       wind_speed, start, end, interval_label,
                                       'bfill')


def gfs_quarter_deg_hourly_to_hourly_mean(latitude, longitude, elevation,
                                          init_time, start, end,
                                          interval_label,
                                          load_forecast=load_forecast,
                                          *, __model='gfs_0p25'):
    """
    Take 1 hr GFS and convert it to hourly average data.
    GHI from NWP model cloud cover. DNI, DHI computed.
    Max forecast horizon 120 hours.

    Parameters
    ----------
    latitude : float
    longitude : float
    elevation : float
    init_time : pd.Timestamp
        Full datetime of a model initialization
    start : pd.Timestamp
        Forecast start. Forecast is inclusive of this instant if
        interval_label is *beginning* and exclusive of this instant if
        interval_label is *ending*.
    end : pd.Timestamp
        Forecast end. Forecast is exclusive of this instant if
        interval_label is *beginning* and inclusive of this instant if
        interval_label is *ending*.
    interval_label : str
        Must be *beginning* or *ending*
    """
    start_floored, end_ceil = _adjust_gfs_start_end(start, end)
    cloud_cover_mixed, air_temperature, wind_speed = load_forecast(
        latitude, longitude, init_time, start_floored, end_ceil, __model,
        variables=('cloud_cover', 'air_temperature', 'wind_speed'))
    cloud_cover = forecast.unmix_intervals(cloud_cover_mixed)
    return _resample_using_cloud_cover(latitude, longitude, elevation,
                                       cloud_cover, air_temperature,
                                       wind_speed, start, end, interval_label,
                                       'bfill')


def gfs_quarter_deg_to_hourly_mean(latitude, longitude, elevation,
                                   init_time, start, end, interval_label,
                                   load_forecast=load_forecast,
                                   *, __model='gfs_0p25'):
    """
    Hourly average forecasts derived from GFS 1, 3, and 12 hr frequency
    output. GHI from NWP model cloud cover. DNI, DHI computed.
    Max forecast horizon 384 hours.

    Parameters
    ----------
    latitude : float
    longitude : float
    elevation : float
    init_time : pd.Timestamp
        Full datetime of a model initialization
    start : pd.Timestamp
        Forecast start. Forecast is inclusive of this instant if
        interval_label is *beginning* and exclusive of this instant if
        interval_label is *ending*.
    end : pd.Timestamp
        Forecast end. Forecast is exclusive of this instant if
        interval_label is *beginning* and inclusive of this instant if
        interval_label is *ending*.
    interval_label : str
        Must be *beginning* or *ending*
    """
    start_floored, end_ceil = _adjust_gfs_start_end(start, end)
    cloud_cover_mixed, air_temperature, wind_speed = load_forecast(
        latitude, longitude, init_time, start_floored, end_ceil, __model,
        variables=('cloud_cover', 'air_temperature', 'wind_speed'))
    cloud_cover = _unmix_various_gfs_intervals(
        init_time, start_floored, end_ceil, cloud_cover_mixed)
    return _resample_using_cloud_cover(latitude, longitude, elevation,
                                       cloud_cover, air_temperature,
                                       wind_speed, start, end, interval_label,
                                       'bfill')


def _adjust_gfs_start_end(start, end):
    """
    Adjusts the GFS start and end times so that we always load a full
    period of the mixed intervals average cycle.
    """
    start_floored = start.floor('6h') + pd.Timedelta('1h')
    if start_floored > start:
        start_floored -= pd.Timedelta('6h')
    end_ceil = end.ceil('6h')
    return start_floored, end_ceil


def _unmix_various_gfs_intervals(init_time, start_floored, end_ceil,
                                 cloud_cover_mixed):
    """unmix intervals for each kind of interval length in GFS forecast"""
    end_1h = init_time + pd.Timedelta('120hr')
    end_3h = init_time + pd.Timedelta('240hr')

    cloud_covers = []

    if start_floored < end_1h:
        cloud_cover_1h_mixed = cloud_cover_mixed.loc[start_floored:end_1h]
        cloud_covers.append(forecast.unmix_intervals(cloud_cover_1h_mixed))

    if end_ceil > end_1h and start_floored < end_3h:
        cloud_cover_3h_mixed = cloud_cover_mixed.loc[
            end_1h+pd.Timedelta('3hr'):end_3h]
        cloud_covers.append(forecast.unmix_intervals(cloud_cover_3h_mixed))

    if end_ceil > end_3h:
        cloud_cover_12h = cloud_cover_mixed.loc[
            end_3h+pd.Timedelta('12hr'):end_ceil]
        cloud_covers.append(cloud_cover_12h)

    cloud_cover = pd.concat(cloud_covers)
    return cloud_cover


def gefs_half_deg_to_hourly_mean(latitude, longitude, elevation,
                                 init_time, start, end, interval_label,
                                 load_forecast=load_forecast,
                                 *, __model='gefs'):
    """
    Hourly average forecasts derived from GEFS 3, 6, and
    12 hr frequency output. GHI from NWP model cloud cover. DNI, DHI
    computed. Max forecast horizon 384 hours.

    Parameters
    ----------
    latitude : float
    longitude : float
    elevation : float
    init_time : pd.Timestamp
        Full datetime of a model initialization
    start : pd.Timestamp
        Forecast start. Forecast is inclusive of this instant if
        interval_label is *beginning* and exclusive of this instant if
        interval_label is *ending*.
    end : pd.Timestamp
        Forecast end. Forecast is exclusive of this instant if
        interval_label is *beginning* and inclusive of this instant if
        interval_label is *ending*.
    interval_label : str
        Must be *beginning* or *ending*

    Returns
    -------
    The columns of the DataFrames correspond to the GEFS members.

    ghi : pd.DataFrame
    dni : pd.DataFrame
    dhi : pd.DataFrame
    air_temperature : pd.DataFrame
    wind_speed : pd.DataFrame
    resample_sort : function
        Resamples, then sorts the above DataFrames.
    solar_position_calculator : function

    Notes
    -----
    Returned values are hourly averages sorted from smallest to largest
    at each time stamp. Each variable is sorted independently.
    This describes a ProbabilisticForecast with ``axis='y'`` and
    ``constant_values=[0, 5, ...95, 100]``.
    """
    start_floored, end_ceil = _adjust_gfs_start_end(start, end)

    def _load_gefs_member(member, solar_position):
        cloud_cover_mixed, air_temperature, wind_speed = load_forecast(
            latitude, longitude, init_time, start_floored, end_ceil, member,
            variables=('cloud_cover', 'air_temperature', 'wind_speed'))
        cloud_cover = _unmix_various_gefs_intervals(
            init_time, start_floored, end_ceil, cloud_cover_mixed)
        return _resample_using_cloud_cover(
            latitude, longitude, elevation, cloud_cover, air_temperature,
            wind_speed, start, end, interval_label, 'bfill',
            solar_position=solar_position)

    # load and process control forecast, then load and process
    # permutations. for efficiency, use control's solar position.
    # accumulate results in dicts so they can be easily converted to
    # DataFrames.
    ghi, dni, dhi, air_temperature, wind_speed, resampler, sol_pos_calc = \
        _load_gefs_member('gefs_c00', None)
    solar_position = sol_pos_calc()
    ghi_ens = {'c00': ghi}
    dni_ens = {'c00': dni}
    dhi_ens = {'c00': dhi}
    air_temperature_ens = {'c00': air_temperature}
    wind_speed_ens = {'c00': wind_speed}
    for member in range(1, 21):
        key = f'p{member:02d}'
        ghi, dni, dhi, air_temperature, wind_speed, _, _ = \
            _load_gefs_member(f'gefs_{key}', solar_position)
        ghi_ens[key] = ghi
        dni_ens[key] = dni
        dhi_ens[key] = dhi
        air_temperature_ens[key] = air_temperature
        wind_speed_ens[key] = wind_speed

    ghi_ens = pd.DataFrame(ghi_ens)
    dni_ens = pd.DataFrame(dni_ens)
    dhi_ens = pd.DataFrame(dhi_ens)
    air_temperature_ens = pd.DataFrame(air_temperature_ens)
    wind_speed_ens = pd.DataFrame(wind_speed_ens)

    def resample_sort(fx):
        resampled = resampler(fx)
        sorted_ = forecast.sort_gefs_frame(resampled)
        return sorted_

    return (ghi_ens, dni_ens, dhi_ens, air_temperature_ens, wind_speed_ens,
            resample_sort, sol_pos_calc)


def _unmix_various_gefs_intervals(init_time, start_floored, end_ceil,
                                  cloud_cover_mixed):
    """unmix intervals for each kind of interval length in GEFS forecast"""
    end_3h = init_time + pd.Timedelta('192hr')

    cloud_covers = []

    if start_floored < end_3h:
        cloud_cover_3h_mixed = cloud_cover_mixed.loc[start_floored:end_3h]
        cloud_covers.append(forecast.unmix_intervals(cloud_cover_3h_mixed))

    if end_ceil > end_3h:
        cloud_cover_6_or_12h = cloud_cover_mixed.loc[
            end_3h+pd.Timedelta('6hr'):end_ceil]
        cloud_covers.append(cloud_cover_6_or_12h)

    cloud_cover = pd.concat(cloud_covers)
    return cloud_cover


def nam_12km_hourly_to_hourly_instantaneous(latitude, longitude, elevation,
                                            init_time, start, end,
                                            interval_label,
                                            load_forecast=load_forecast,
                                            *, __model='nam_12km'):
    """
    Hourly instantantaneous forecast.
    GHI directly from NWP model. DNI, DHI computed.
    Max forecast horizon 36 hours.

    Parameters
    ----------
    latitude : float
    longitude : float
    elevation : float
    init_time : pd.Timestamp
        Full datetime of a model initialization
    start : pd.Timestamp
        Forecast start. Forecast is inclusive of this point.
    end : pd.Timestamp
        Forecast end. Forecast is exclusive of this point.
    interval_label : str
        Must be instant
    """
    start_adj, end_adj = adjust_start_end_for_interval_label(
        interval_label, start, end, limit_instant=True)
    ghi, air_temperature, wind_speed = load_forecast(
        latitude, longitude, init_time, start_adj, end_adj, __model,
        variables=('ghi', 'air_temperature', 'wind_speed'))
    dni, dhi, solar_pos_calculator = _ghi_to_dni_dhi(
        latitude, longitude, elevation, ghi)
    # hourly instant in, hourly instant out
    resampler = partial(forecast.resample, freq='1h')
    return (ghi, dni, dhi, air_temperature, wind_speed,
            resampler, solar_pos_calculator)


def nam_12km_cloud_cover_to_hourly_mean(latitude, longitude, elevation,
                                        init_time, start, end, interval_label,
                                        load_forecast=load_forecast,
                                        *, __model='nam_12km'):
    """
    Hourly average forecast.
    GHI from NWP model cloud cover. DNI, DHI computed.
    Max forecast horizon 72 hours.

    Parameters
    ----------
    latitude : float
    longitude : float
    elevation : float
    init_time : pd.Timestamp
        Full datetime of a model initialization
    start : pd.Timestamp
        Forecast start. Forecast is inclusive of this instant if
        interval_label is *beginning* and exclusive of this instant if
        interval_label is *ending*.
    end : pd.Timestamp
        Forecast end. Forecast is exclusive of this instant if
        interval_label is *beginning* and inclusive of this instant if
        interval_label is *ending*.
    interval_label : str
        Must be *beginning* or *ending*
    """
    cloud_cover, air_temperature, wind_speed = load_forecast(
        latitude, longitude, init_time, start, end, __model,
        variables=('cloud_cover', 'air_temperature', 'wind_speed'))
    return _resample_using_cloud_cover(latitude, longitude, elevation,
                                       cloud_cover, air_temperature,
                                       wind_speed, start, end, interval_label,
                                       'interpolate')
